import sys
# sys.path.append(r'C:\Users\lass_j\Documents\Software\DMCpy')
from Tutorial_Class import Tutorial
import os
# import numpy as np

def Tester():
    from DMCpy import DataSet
        
    # print general help for exporting data.
    DataSet.export_help()
    
    # print help for add() function. 
    help(DataSet.add)    
    

    # export(): Exports 565 and 578 induyvidually. The step size for the exported files is 0.25. The data files are located in 'data/' and the exported files are stored in 'docs/Tutorials/Powder'. 
    DataSet.export(565,578,dTheta=0.25,folder=r'data/',outFolder=r'docs/Tutorials/Powder')
    # exports .dat and .xye files of 565 and 578 induvidually.

    # export(): Export can also be used to merge files. Here [567,568,570,571] is merged, '570-573' is merged and (574,575) is merged.
    # In the file names of the exported files, the file numbers are given, and not the sample name. 
    DataSet.export([567,568,570,571],'570-573',(574,575),sampleName=False,fileNumber=True,folder=r'data/',outFolder=r'docs/Tutorials/Powder')
    # exports .dat and .xye files of 567_568_570_571, 570-573, 574_575
    
    # add(): Add the files 565,578,579,585,586,587,575 and export one file named 'added_files'. The data files are located in 'data/' and the exported files are stored in 'docs/Tutorials/Powder'. 
    DataSet.add(565,578,579,(585),'586-587',[575],outFile='added_files',folder=r'data/',outFolder=r'docs/Tutorials/Powder')
    # exports 'added_files.dat' and 'added_files.xye'
    
    # export_from_to(): exports all files from 578 to 582. The files are located in 'data/' and the exported files are stored in 'docs/Tutorials/Powder'. 
    # For the automatic filename, sample name is not included, but the file number is included. 
    DataSet.export_from_to(578,582,sampleName=False,fileNumber=True,folder=r'data/',outFolder=r'docs/Tutorials/Powder')
    # exports .dat and .xye files of 578, 579, 580, 581, 582
    
    # subtract(): Subtract two data files from each other. Must have same binning. In this case, only .dat files are subtracted as xye=False. 
    DataSet.subtract('DMC_579','DMC_578',xye=False,outFile=r'subtracted_file',folder=r'docs/Tutorials/Powder',outFolder=r'docs/Tutorials/Powder')
    # create subtracted_file.dat
    
    
title = 'Export of diffraction patterns'

introText = 'The software can export powder patterns to .dat files in PSI format and to .xye files. This is done by built in functions. '\
+'The following fuctions are avaliable: *export()*, *add()*, *export_from()*, *export_from_to()* to export measured data to . '\
+'In addition, the function *export_help()* can be used to general help for export functions. '\
+'Help for all export function can also be printed by e.g. *help(export)*. '\
+'In this tutorial we examplify the use of the export functions and the various keywords for the functions.\n \n'\
+'Properties of export functions:\n'\
+' - export(): For export of induvidual sets of scan files. Files can be merged by [] or "" notation, i.e. list or strings.\n'\
+' - add(): The function adds/merge all the files given independent on the notation.\n'\
+' - export_from(): For export of all data files in a folder after a startfile.\n'\
+' - export_from_to(): It exports all files between and including two given file numbers.\n'\
+' - export_list(): Takes a list and export all the files separatly. If a list is given inside the list, the files will be merged.\n'\
+' - subtract(): Takes two file names and subtract the corresponding .dat and .xye files. The function act on exported data files, not on raw data.\n \n'\
+'Most important kewords:\n'\
+' - sampleName (bool): Include sample name from raw file in filename. Default is True.\n'\
+' - dTheta (float): Stepsize of binning if no bins is given (default is 0.125).\n'\
+' - outFile (str): String for name of outfile (given without extension). If not given, an automatic file name will be generated.\n'\
+' - folder (str): Path to directory for data files, default is current working directory.\n'\
+' - outFolder (str): Path to folder data will be saved. Default is current working directory.\n \n'\
+'Example: export(578, dTheta=0.25)\n \n'\
+'Both PSI and xye format files are exported as default. They can be deactivated by the following keywords.\n'\
+' - PSI (bool): Use PSI format. Default is True.\n'\
+' - xye (bool): Use xye format. Default is True.\n \n'\
+'Example: export(578, PSI=False) \n \n'\
+'File names for exported files are by default the sample named stored in the raw file. This can be changed with the following keywords:\n \n'\
+' - sampleName (bool): Use sample name stored in raw file in file name of exported data. Default is True.\n'\
+' - temperature (bool): Include temperature in filename. Default is False.\n'\
+' - fileNumber (bool): Include sample number in filename. Default is False.\n'\
+' - magneticField (bool): Include magnetic field in filename. Default is False.\n'\
+' - electricField (bool): Include electric field in filename. Default is False.\n \n'\
+'Example: export(578, sampleName=False, fileNumber=True)\n \n'\
+'Examples fo use of export functions: \n \n'


outroText = ' '
    

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('ExportPowder',introText,outroText,Tester,fileLocation = (os.path.join(os.getcwd(),r'docs/Tutorials/Powder'))) 

def test_ExportPowder():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()


import os.path
import numpy as np
from DMCpy import _tools



def test_fileListGenerator():

    fileString = '10-12,15-21,23'

    folder = 'data'
    year = 2021
    fileList = _tools.fileListGenerator(fileString,folder=folder,year=year)

    expected = [os.path.join('data',f) for f in ['dmc2021n000010.hdf', 'dmc2021n000011.hdf', 'dmc2021n000012.hdf', 'dmc2021n000015.hdf', 'dmc2021n000016.hdf', 'dmc2021n000017.hdf', 'dmc2021n000018.hdf', 'dmc2021n000019.hdf', 'dmc2021n000020.hdf', 'dmc2021n000021.hdf', 'dmc2021n000023.hdf']]  
    assert(np.all(expected==fileList))

    # Generate year and fileString containing data file numbers from fileList
    yearGenerated,fileStringGenerated = _tools.numberStringGenerator(fileList)

    assert(yearGenerated==year)
    assert(fileStringGenerated == fileString)


def test_roundPower():
    assert(_tools.roundPower(0.0,default=10) == 10) # Default is returnd at 0

    assert(_tools.roundPower(0.1) == 1)

    assert(_tools.roundPower(0.01) == 2)

    assert(_tools.roundPower(0.09) == 2)

    assert(_tools.roundPower(10.09) == -1)

    assert(_tools.roundPower(1.09) == 0)


import sys,os
sys.path.append('.')
import pickle,numpy as np

__version__ = '0.2.1'
__author__ = 'Jakob Lass'

# installFolder = os.path.abspath(os.path.split(__file__)[0])
# calibrationFile =  os.path.join(installFolder,'calibrationDict.dat')

installFolder = os.path.abspath(os.path.join(os.path.split(__file__)[0],'..'))
calibrationFile =  os.path.join(installFolder,'DMCpy','calibrationDict.dat')

try:
    with open(calibrationFile,'rb') as f:
        calibrationDict = pickle.load(f)
except FileNotFoundError:
    def find(name, path):
        result = []
        for root, dirs, files in os.walk(path):
            if name in files:
                result.append(os.path.join(root, name))
        return result

    
    foundFile = str(find('calibrationDict.dat',os.path.abspath(os.path.join(installFolder,'..','..','..','..','..','..')))[0])

    with open(foundFile,'rb') as f:
        calibrationDict = pickle.load(f)
    #raise FileNotFoundError

    
calib2021 = {'limits':np.array([1]),
             'names':np.array(['Mockup']),
             'Mockup':np.ones((1152,128))}

calibrationDict[2021] = calib2021


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 14 11:54:33 2021

@author: Jakob Lass

Tool to replicate behaviour of make file
"""

import argparse
from ntpath import join

import os,glob

possibleArguments = ['test', 'tutorials', 'wheel', 'html', 'upload', 'testVersion', 'version']

parser = argparse.ArgumentParser(description="Make tool replacing linux Makefile.")
parser.add_argument("task", nargs='?', default='help', type=str, help="Type of task to be performed. Possible tasks are: {}. Run without argument to see help menu.".format(', '.join(possibleArguments)))
parser.add_argument("version", nargs='?',default=None, type=str, help="Version number to be created. Default is previous plus 1, i.e. x.y.z -> x.y.z+1")

def callHelp(parser=parser):
    parser.print_help()

def test():
    os.system("python -m pytest -vv test")

def cleanTutorialFolders():
    folders = [f.path for f in os.scandir(os.path.join('docs','Tutorials')) if f.is_dir()]
    print('Cleaning tutorial folders: '+', '.join(folders))
    for folder in folders:
        os.system('python clean.py '+folder)

def generateTutorials():
    os.system('python '+str(os.path.join('Tutorials','tutorials.py')))
    

def makeHTML():
    print('Creating HTML through Sphinx')
    os.system("sphinx-build docs build")

def makeWheel():
    os.system("python setup.py sdist")

def getLatestBuild():
    list_of_files = glob.glob(os.path.join('dist','*')) # * means all if need specific format then *.csv
    return max(list_of_files, key=os.path.getctime)

def getCurrentVersion():
    string  = getLatestBuild()
    versionList = string.split('-')[1].split('.')[:-2]
    if 'post' in versionList[-1]:
        versionList[-1] = versionList[-1].replace('post','')
        adder = 'post'
    else:
        adder = ''
    version = [int(x) for x in versionList]
    
    if adder != '':
        version[-1]=adder+str(version[-1]+1)
    else:
        version[-1]+=1
    return '.'.join([str(x) for x in version])

def uploadPyPi(server='testpypi'):
    build = getLatestBuild()
    print('Uploading ' + str(build) + ' to '+server)
    os.system('twine upload {} -r {}'.format(build,server))


def upload():
    uploadPyPi(server='testpypi')
    uploadPyPi(server='pypi')

def update(version):
    os.system('python Update.py '+version)

def makeTutorials():
    cleanTutorialFolders()
    generateTutorials()
    makeHTML()

args = parser.parse_args()


if args.task == '' or args.task.lower() == 'help':
    callHelp()
elif args.task.lower() == 'test':
    test()

elif args.task.lower() == 'tutorials':
    makeTutorials()

elif args.task.lower() == 'wheel':
    makeWheel()

elif args.task.lower() == 'upload':
    upload()

elif args.task.lower() == 'testversion':
    if args.version is None:
        version = getCurrentVersion()
    else:
        version = args.version
    print('Creating test version '+version)
    update(version=version)
    makeWheel()
    uploadPyPi(server='testpypi')

elif args.task.lower() == 'html':
    makeHTML()

elif args.task.lower() == 'version':
    if args.version is None:
        version = getCurrentVersion()
    else:
        version = args.version
    print('Creating version '+version)
    update(version=version)
    makeTutorials()
    addFiles = ['setup.py',
                os.path.join('docs','conf.py'),
                os.path.join('docs','Tutorials','*'),
                os.path.join('docs','index.rst'),
                os.path.join('DMCpy','__init__.py'),
                os.path.join('test','init.py'),
                'setup.py',
                ]
    os.system("git add {}".format(' '.join(addFiles)))
    os.system("git commit -m \"Update version\"")
    os.system("git tag -a {0} -m \'{0}\'".format(version))
    makeWheel()
    os.system("git push")
    os.system("git push --tags")

else:
    callHelp()


import argparse

OpossibleArguments = ['Nope','None','I don\'t know']

parser = argparse.ArgumentParser(description="Collection of DMCpy tools for the command line")
#parser.add_argument("task", nargs='?', default='help', type=str, help="Type of task to be performed. Possible tasks are: {}. Run without argument to see help menu.".format(OpossibleArguments))


args = parser.parse_args()

print(args)



from cmath import acos
import numpy as np
from DMCpy import _tools
import h5py as hdf
from DMCpy import TasUBlibDEG


def cosd(x):
    return np.cos(np.deg2rad(x))

def sind(x):
    return np.sin(np.deg2rad(x))

def camelCase(string,split='_'):
    """Convert string to camel case from <split> seperated"""

    if not split in string:
        return string
    splitString = string.split(split)
    first = splitString[0]
    others = [x.title() for x in splitString[1:]]
    
    combi = [first]+others
    return ''.join([str(x) for x in combi])

class Sample(object):
    """Sample object to store all information of the sample from the experiment"""
    @_tools.KwargChecker()
    def __init__(self,a=1.0,b=1.0,c=1.0,alpha=90,beta=90,gamma=90,sample=None,name='Unknown',projectionVector1=None, projectionVector2 = None):

        
        if isinstance(sample,hdf._hl.group.Group):
            self.name = np.array(sample.get('name'))[0].decode()
            if self.name is None or self.name == '':
                self.name = 'Unknown'

            unitCell = np.array(sample.get('unit_cell'))

            if not sample.get('unit_cell') is None:
                self.unitCell = unitCell
            else:
                self.unitCell = [1,1,1,90,90,90]
            self.UB = self.B
            
            
        elif np.all([a is not None,b is not None, c is not None]):
            self.unitCell = np.array([a,b,c,alpha,beta,gamma])
           
            self.polarAngle = np.array(None)
            self.rotationAngle = np.array(0)
            self.name=name
            
            r1 = projectionVector1
            r2 = projectionVector2
            self.plane_vector1 = r1
            self.plane_vector2 = r2

            self.planeNormal = np.cross(self.plane_vector1[:3],self.plane_vector2[:3])
            self.UB = self.B
            
            # cell = TasUBlib.calcCell(self.unitCell)

            # self.orientationMatrix = TasUBlib.calcTasUBFromTwoReflections(cell, r1, r2)
            # #self.orientationMatrix = TasUBlib.calcTasUBFromTwoReflections(self.cell,self.plane_vector1,self.plane_vector2)
            # self.projectionVector1,self.projectionVector2 = calcProjectionVectors(self.plane_vector1.astype(float),self.plane_vector2.astype(float))#,self.planeNormal.astype(float))
            # self.initialize()
            # self.calculateProjections()
        else:
            print(sample)
            print(a,b,c,alpha,beta,gamma)
            raise AttributeError('Sample not understood')

            
    @property
    def unitCell(self):
        return self._unitCell

    @unitCell.getter
    def unitCell(self):
        return np.array([self.a,self.b,self.c,self.alpha,self.beta,self.gamma])#self._unitCell

    @unitCell.setter
    def unitCell(self,unitCell):
        self._unitCell = unitCell
        self.a = unitCell[0]
        self.b = unitCell[1]
        self.c = unitCell[2]
        self.alpha = unitCell[3]
        self.beta  = unitCell[4]
        self.gamma = unitCell[5]
        self.updateCell()
        
    @property
    def a(self):
        return self._a

    @a.getter
    def a(self):
        return self._a

    @a.setter
    def a(self,a):
        if a>0:
            self._a = a
        else:
            raise AttributeError('Negative or null given for lattice parameter a')

    @property
    def b(self):
        return self._b

    @b.getter
    def b(self):
        return self._b

    @b.setter
    def b(self,b):
        if b>0:
            self._b = b
        else:
            raise AttributeError('Negative or null given for lattice parameter b')

    @property
    def c(self):
        return self._c

    @c.getter
    def c(self):
        return self._c

    @c.setter
    def c(self,c):
        if c>0:
            self._c = c
        else:
            raise AttributeError('Negative or null given for lattice parameter c')


    @property
    def alpha(self):
        return self._alpha

    @alpha.getter
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self,alpha):
        if alpha>0 and alpha<180:
            self._alpha = alpha
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter alpha')

    @property
    def beta(self):
        return self._beta

    @beta.getter
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self,beta):
        if beta>0 and beta<180:
            self._beta = beta
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter beta')

    @property
    def gamma(self):
        return self._gamma

    @gamma.getter
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self,gamma):
        if gamma>0 and gamma<180:
            self._gamma = gamma
        else:
            raise AttributeError('Negative,null or above 180 degrees given for lattice parameter gamma')

    @property
    def UBInv(self):
        return np.linalg.inv(self.UB)

    # @property
    # def convertinv(self):
    #     return np.linalg.inv(self.convert)

    # @property
    # def projectionAngle(self):
    #     p1 = np.dot(self.convertHKL,self.projectionVectors[0])
    #     p2 = np.dot(self.convertHKL,self.projectionVectors[1])
    #     return np.rad2deg(_tools.vectorAngle(p1,p2))

    def updateCell(self):
        self.fullCell = TasUBlibDEG.calcCell(self.unitCell)
        self.B = TasUBlibDEG.calculateBMatrix(self.fullCell)

    def saveToHdf(self,entry):
        entry.create_dataset('name',data = [np.string_(self.name)])
        if hasattr(self,'unitCell'):
            entry.create_dataset('unit_cell',data = self.unitCell)

        
    def defineUB(self,HKL1,HKL2,Q1,Q2):
        self.P1 = _tools.LengthOrder(HKL1)
        self.P3 = _tools.LengthOrder(np.cross(HKL1,HKL2))
        self.P2 = _tools.LengthOrder(np.cross(self.P3,HKL1))

        self.projectionVectors = np.array([self.P1,self.P2,self.P3]).T

        # projectionVectors = np.array([[0,0,1],[1,1,0],[1,-1,0]])
        axisVectors = np.eye(3)
        ## Assume that Q1/HKL1 is along x-axis

        Alpha1 = np.rad2deg(np.arccos(np.dot(Q1,axisVectors[0])/(np.linalg.norm(Q1))))
        Rot1 = np.cross(Q1,axisVectors[0])
        Rot1*=1.0/np.linalg.norm(Rot1)
        ROT1 = _tools.rotMatrix(Rot1,Alpha1)

        # Rotate Q2 into Q1's frame
        Q2Rot = np.dot(ROT1,Q2)
        Q2Rot-= np.dot(axisVectors[0],Q2Rot)*axisVectors[0]# project out [1,0,0] as this rotation has been done by Q1

        Alpha2 = np.rad2deg(np.arccos(np.dot(Q2Rot,axisVectors[1])/(np.linalg.norm(Q2Rot))))#np.rad2deg(np.arccos(Q2Rot[1]/np.linalg.norm(Q2Rot)))
        Rot2 = np.cross(Q2Rot,axisVectors[1])
        Rot2*=1.0/np.linalg.norm(Rot2)
        ROT2 = _tools.rotMatrix(Rot2,Alpha2)

        self.ROT = np.dot(ROT2,ROT1)

        
        self.projectionB = np.diag(np.linalg.norm(np.dot(self.projectionVectors.T,self.B),axis=1))

        # Rotates into the scattering plane
        self.UB = np.dot(self.ROT.T,np.dot(self.projectionB,np.linalg.inv(self.projectionVectors)))#np.linalg.inv(np.dot(Binverse,self.ROT))

        # p23 = np.array([[1,0,0],[0,1,0]]) 
        # self.convert = np.dot(p23,np.einsum('ij,jk->ik',self.UB,self.projectionVectors[:,:2])) # Convert from projX,projY,projZ to Qx, Qy, Qz
        # self.convertHKL = np.dot(p23,self.UB) # Convert from HKL to Qx, Qy
        
    def tr(self,proj0,proj1,proj2=None,projection=2):
        """Convert from projX, projY coordinate to Qx',QY' coordinate."""
        if proj2 is None:
            p0, p1 = np.asarray(proj0), np.asarray(proj1)
            P = np.array([p0,p1])

            projections = np.delete(self.projectionVectors,projection,axis=1)
            
            pm = np.delete(np.eye(3),projection,axis=0)
            
            convert = np.dot(pm,np.dot(self.ROT,np.dot(self.UB,projections)))
        else:

            p0, p1, p2 = np.asarray(proj0), np.asarray(proj1), np.asarray(proj2)
            P = np.array([p0,p1,p2])

            # permutation order of the projection vectors
            order = np.array([[1,2,0],[0,2,1],[0,1,2]])

            projections = self.projectionVectors[:,order[projection]]#np.delete(self.projectionVectors,projection,axis=1)
            #projections = np.concatenate([projections,np.array([self.projectionVectors[:,projection]]).reshape(3,1)],axis=1)
            
            convert = np.dot(self.ROT,np.dot(self.UB,projections))
        return np.einsum('ij,j...->i...',convert,P)


    def inv_tr(self,qx,qy, qz = None,projection=2):
        """Convert from projX, projY coordinate to Qx',QY' coordinate."""
        if qz is None:

            p0, p1 = np.asarray(qx), np.asarray(qy)
            P = np.array([p0,p1])
            projections = np.delete(self.projectionVectors,projection,axis=1)
            
            pm = np.delete(np.eye(3),projection,axis=0)

            convert = np.linalg.inv(np.dot(pm,np.dot(self.ROT,np.dot(self.UB,projections))))
        else:
            p0, p1, p2 = np.asarray(qx), np.asarray(qy), np.asarray(qz)
            P = np.array([p0,p1,p2])

            # permutation order of the projection vectors
            order = np.array([[1,2,0],[0,2,1],[0,1,2]])

            projections = self.projectionVectors[:,order[projection]]#np.delete(self.projectionVectors,projection,axis=1)
            #projections = np.concatenate([projections,np.array([self.projectionVectors[:,projection]]).reshape(3,1)],axis=1)
            
            convert = np.linalg.inv(np.dot(self.ROT,np.dot(self.UB,projections)))
        
        
        return np.einsum('ij,j...->i...',convert,P)


    def projectionAngle(self,projection=2):
        V1 = self.tr(1,0,projection=projection)
        V2 = self.tr(0,1,projection=projection)
        return _tools.vectorAngle(V1, V2)

    def format_coord(self,x,y,z=None,projection=2):
        """Format coordinates from Qx'Qy' in rotated frame into HKL."""
        order = np.array([[1,2,0],[0,2,1],[0,1,2]])
        if z is None:
            proj0,proj1 = self.inv_tr(x,y,projection=projection)
            projections = np.delete(self.projectionVectors,projection,axis=1)
            rlu = proj0*projections[:,0]+proj1*projections[:,1]
        else:
            proj0,proj1,proj2 = self.inv_tr(x,y,z,projection=projection)
            projections = self.projectionVectors[:,order[projection]]
            
            rlu = proj0*projections[:,0]+proj1*projections[:,1]+proj2*projections[:,2]
        return "h = {0:.3f}, k = {1:.3f}, l = {2:.3f}".format(rlu[0],rlu[1],rlu[2])

    
    def calculateQxQyQzToHKL(self,x,y,z):
        """convert from Qx,Qy to HKL."""
        pos = np.array([x,y,z])
        return np.einsum('ij,j...->i...',self.UBInv,pos)

    def calculateHKLToQxQyQz(self,H,K,L):
        """convert HKL to Qx,Qy."""
        pos = np.array([H,K,L])
        return np.einsum('ij,j...->i...',self.UB,pos)

    def calculateHKLtoProjection(self,H,K,L):
        """convert from projections to HKL."""
        #QxQyQz = np.dot(self.ROT,self.calculateHKLToQxQyQz(H,K,L))
        #projection = self.inv_tr(*QxQyQz)
        projection = np.dot(np.linalg.inv(self.projectionVectors),np.array([H,K,L]))
        return projection


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Wed Jul 14 13:03:12 2021

@author: Jakob Lass

Tool to replace batch in order to fully port development to windows
"""

import os,sys,shutil

DIR = str(os.path.dirname(__file__))

try:
    os.remove(os.path.join('Tutorials','tutorialList.txt'))
except FileNotFoundError:
    pass

files = [f.path for f in os.scandir(DIR) if f.is_file()]

for f in files:
    if not f[-2:] == 'py':
        continue
    if 'tutorials.py' in f:
        continue
    os.system('python '+f)

    
    

for folder in ['.cache','.pyteste_cache','__pycache']:
    try:
        shutil.rmtree(os.path.join(DIR,folder))
    except FileNotFoundError:
        pass


# Generate correct tutorials.rst file
tutorialFile = os.path.join('docs','Tutorials','Tutorials.rst')

mainText = """.. _Tutorials:

Tutorials
---------
Main text about the tutorials below..

.. toctree::
   :maxdepth: 1
"""

with open(os.path.join('Tutorials','tutorialList.txt'),'r') as f:
    tutorials = f.readlines()
            
tutorialText = []
for tut in tutorials:
    common = os.path.commonpath([tut,os.path.abspath(tutorialFile)]).strip()
    tutorialText.append(str(os.path.relpath(tut,common)).replace('\\','/'))


text = mainText+'\n   '+'\n   '.join(tutorialText)+'\n\n'



with open(tutorialFile,'w') as f:
    f.write(text)




import sys
# sys.path.append(r'C:\Users\lass_j\Documents\Software\DMCpy')
from Tutorial_Class import Tutorial
import os

def Tester():
    from DMCpy import DataSet,DataFile, _tools
    
    # To plot powder data we give the file number, or list of filenumbers as a string and the folder of the raw data
    scanNumbers = '578'
    folder = 'data'
    
    # In this case, we have a offset of twoTheta of 18 degrees. We correct this by changeing the twoThetaOffset for all dataFiles
    twoThetaOffset = 18.0
    
    # a list of dataFiles are generated with loadDataFile running over all the dataFiles generated from _tools.fileListGenerator and twoThetaOffset acts on the dataFile
    dataFiles = [DataFile.loadDataFile(dFP,twoThetaPosition=twoThetaOffset) for dFP in _tools.fileListGenerator(scanNumbers,folder)]

    # We then create a data set based on the data files
    ds = DataSet.DataSet(dataFiles)
        
    # We can also give the step size for the integration. Default is 0.2 
    dTheta = 0.125
    
    # Generate a diffraction pattern where the 2D detector is integrated in Q-space
    ax,bins,Int,Int_err,monitor = ds.plotTwoTheta(dTheta=dTheta)
    ax.set_title('Integrated in Q')
    fig = ax.get_figure()
    fig.savefig(os.path.join(os.getcwd(),r'docs/Tutorials/Powder/TwoThetaPowderQ.png'),format='png',dpi=300)
    
    # Generate a diffraction pattern where the 2D detector is integrated vertically
    ax2,bins2,Int2,Int_err2,monitor2 = ds.plotTwoTheta(correctedTwoTheta=False,dTheta=dTheta)
    ax2.set_title('Integrated vertically')
    fig2 = ax2.get_figure()
    fig2.savefig(os.path.join(os.getcwd(),r'docs/Tutorials/Powder/TwoThetaPowderVertical.png'),format='png',dpi=300)
 
    
title = 'Plot of diffraction patterns'

introText = 'When a powder sample has been measured at DMC it is saved in hdf files. Several DataFiles can be combined into a common DataSet and plotted. '\
+'The follwing code takes a DatsSet, here consisting of one single DataFile, and plot it. Two different settings '\
+'for the binning method is used *correctedTwoTheta* equal to *True* and *False*. When *False* a naive summation across the '\
+'2D detector is performed where the out-of-plane component is not taken into account. That is, summation is performed vertically '\
+'on the detector. For powder patterns around 90\ :sup:`o`, this is only a very minor error, but for scattering close to the direct '\
+'beam a significant error is introduced. Instead, utilizing *correctedTwoTheta = True* is the correct way. The scattering 3D '\
+'vector is calculated for each individual pixel on the 2D detector and it\'s length is calculated.'
   


outroText = 'Running the above code generates the two following, similar looking, diffractograms utilizing the corrected and uncorrected '\
+'twoTheta positions respectively. We highlight the reduced noice and the slightly sharper peaks in the corrected image.'\
+'The down-up glitches of the diffraction pattern is from the interface between the detector sections.'\
+'\n .. figure:: TwoThetaPowderQ.png\n  :width: 50%\n  :align: center\n\n'\
+'\n .. figure:: TwoThetaPowderVertical.png\n  :width: 50%\n  :align: center\n\n'    

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('Powder Diffractogram',introText,outroText,Tester,fileLocation = (os.path.join(os.getcwd(),r'docs/Tutorials/Powder')))

def test_Powder_Diffractogram():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()
    
    
    
 # ds = DataSet.generate_data_set(565,folder=r'data',twoThetaOffset=18)   
    


from DMCpy import DataFile,_tools
import os.path
import numpy as np
import matplotlib.pyplot as plt

def test_init():
    df = DataFile.DataFile()

    try:
        df = DataFile.DataFile(r'Wrong\Path') # File not found
        assert False
    except FileNotFoundError:
        assert True


    df = DataFile.loadDataFile(fileLocation=os.path.join('data','dmc2021n000565.hdf'))
    path,name = os.path.split(os.path.join('data','dmc2021n000565.hdf'))

    assert(df.folder == path)
    assert(df.fileName == name)
    assert(df.fileType.lower() == 'powder')


def test_copy(): # Test the ability to copy from one data file to another
    testDF = DataFile.DataFile(os.path.join('data','dmc2021n000565.hdf'))

    dfCopy = DataFile.loadDataFile(testDF) # Perform copy

    assert(dfCopy==testDF)
    
def test_load():
    testDF = DataFile.loadDataFile(os.path.join('data','dmc2021n000565.hdf'))

    assert(testDF.twoTheta.shape == (128,128*9))
    assert(testDF.counts.shape == (1,128,128*9))
    assert(testDF.correctedTwoTheta.shape == (1,128,128*9))

    # If detector is assumed to be flat, twoTheta and correctedTwoTheta are the same at middle
    
    assert(np.all(np.isclose(np.mean(testDF.correctedTwoTheta[0,[63,64],:],axis=0),testDF.twoTheta[63,:],atol=0.06)))

    #testDF = DataFile.DataFile(os.path.join('data','dmc2018n000401 - Copy.hdf'))

    #assert(testDF.twoTheta.shape == (400,100))
    #assert(testDF.counts.shape == (400,100))
    #assert(testDF.correctedTwoTheta.shape == (400,100))

    # If detector is assumed to be flat, twoTheta and correctedTwoTheta are the same
    

def test_plot():
    dataFile = os.path.join('data','dmc2021n{:06d}.hdf'.format(494))

    df = DataFile.loadDataFile(dataFile)
    fig,ax = plt.subplots()

    Ax = df.plotDetector()


def test_masking_2D():
    df = DataFile.loadDataFile()

    # An empty data file raises error on making a mask
    try:
        df.generateMask()
        assert False
    except RuntimeError:
        assert True

    df = DataFile.loadDataFile(os.path.join('data','dmc2021n{:06d}.hdf'.format(494)))

    df.generateMask(maxAngle=90) # No points are masked
    assert(np.all(df.mask==np.zeros_like(df.counts,dtype=bool)))

    df.generateMask(maxAngle=-1) # All points are masked
    assert(np.all(df.mask==np.ones_like(df.counts,dtype=bool)))

    df.generateMask(maxAngle=7) # All points are masked
    total = np.size(df.counts)
    maskTotal = np.sum(np.logical_not(df.mask))
    assert(total>maskTotal)

    try:
        df.generateMask(MaxAngle=7)
        assert(False)
    except AttributeError as e:
        assert(e.args[0] == 'Key-word argument "MaxAngle" not understood. Did you mean "maxAngle"?')

    try:
        ax = df.plotDetector(aplynormalization=True)
        assert(False)
    except AttributeError as e:
        assert(e.args[0] == 'Key-word argument "aplynormalization" not understood. Did you mean "applyNormalization"?')
        


def test_calibration():
    fileName = 'dmc2018n000250.hdf' # no calibration exists

    calibData,calibName = DataFile.findCalibration(fileName)
    assert(calibName == 'None')
    assert(calibData is None)

    fileName = 'dmc2018n036099.hdf' # calibration deteff_18c.dat

    calibData,calibName = DataFile.findCalibration(fileName)
    assert(calibName == 'deteff_18c.dat')
    assert(calibData.shape == (400,))

    # Test when full data file path is provided
    fileName = os.path.join("fictive","folder","to","data",'dmc2002n000099.hdf') # calibration deteff_02c.dat
    calibData,calibName = DataFile.findCalibration(fileName)
    assert(calibName == 'deteff_02c.dat')
    assert(calibData.shape == (400,))

    # year not covered in calibration data
    fileName = 'dmc2019n000250.hdf'
    try:
        calibData,calibName = DataFile.findCalibration(fileName)
        assert(False)
    except FileNotFoundError:
        assert True


def test_decoding():
    dataFile = os.path.join('data','dmc2021n{:06d}.hdf'.format(494))

    df = DataFile.loadDataFile(dataFile)

    assert(isinstance(df.sample.name,str)) # Originally byte array
    

def test_saveLoad():

    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(number)) for number in [565,597]]
    for dataFile in dataFiles:
        df = DataFile.loadDataFile(dataFile)

        splitted = dataFile.split('.')
        splitted[-2] = splitted[-2] +'new'
        saveFileName = '.'.join(splitted)


        if os.path.exists(saveFileName):
            os.remove(saveFileName)
            
        df.save(saveFileName)
        df2 = DataFile.loadDataFile(saveFileName)
        assert(df==df)
        assert(df==df2)

        if os.path.exists(saveFileName):
            os.remove(saveFileName)


def test_changeOfParameters():
    dataFile = os.path.join('data','dmc2021n{:06d}.hdf'.format(494))
    df = DataFile.loadDataFile(dataFile,twoThetaPosition=np.array([1])) # Move the two theta position away from absolute 0

    originalKi = df.Ki
    originalWaveLength = df.wavelength
    originalQ = df.q

    df.Ki = 2.0
    assert(np.isclose(df.wavelength,np.pi))
    assert(np.all(np.logical_not(np.isclose(df.q,originalQ))))

    df.wavelength = 2.0
    assert(np.isclose(df.Ki,np.pi))
    assert(np.all(np.logical_not(np.isclose(df.q,originalQ))))
    
    df.Ki = originalKi
    assert(np.isclose(df.wavelength,originalWaveLength))
    assert(np.all(np.isclose(df.q,originalQ)))

    df2 = DataFile.loadDataFile(dataFile)
    df2.twoThetaPosition = np.array([1])
    assert(df==df2)


    splitted = dataFile.split('.')
    splitted[-2] = splitted[-2] +'new'
    saveFileName = '.'.join(splitted)
    if os.path.exists(saveFileName):
        os.remove(saveFileName)
        
    ## Change all parameters and save to check reproducibility
    df.Ki = 2
    df.twoThetaPosition = np.array([20.0])
    
    df.save(saveFileName)
    df2 = DataFile.loadDataFile(saveFileName)
    assert(df==df2)

    if os.path.exists(saveFileName):
        os.remove(saveFileName)
    


def test_shallow_read():
    parameters = ['startTime','twoThetaPosition','wavelength','sampleName']

    files = _tools.fileListGenerator('494,565',folder=r'data',year=2021)

    dicts = DataFile.shallowRead(files,parameters)

    startTimes = ['2021-12-17 15:14:59','2021-12-21 17:27:35']
    sampleNames = ['','']

    wavelength = 2.4500992

    for I,(file,d) in enumerate(zip(files,dicts)):
        assert(d['file'] == file)
        assert(np.isclose(d['wavelength'],wavelength))
        assert(startTimes[I] == d['startTime'])
        assert(d['sampleName'] == sampleNames[I])
        




import copy
import functools
import os
import sys
from operator import sub

pythonVersion = sys.version_info[0]

sys.path.append('.')
sys.path.append('..')
sys.path.append('../..')

import matplotlib.pyplot as plt
#if pythonVersion == 3: # Only for python 3
import matplotlib.ticker as mticker

import numpy as np
from DMCpy import _tools
from mpl_toolkits.axisartist import SubplotHost
try:
    from mpl_toolkits.axisartist.grid_helper_curvelinear import \
        GridHelperCurveLinear
except ImportError:
    from mpl_toolkits.axisartist.floating_axes.grid_helper_curvelinear import \
        GridHelperCurveLinear



def calculateBase(l,span,ticks):
    """Calcualte the tick mark base suitable for current span and ticks

    Args:

        - l (grid locator): Matplotlib grid locator

        - span (float): Width of view

        - ticks (int): Number of ticks wanted

    Returns:

        - base (float): Closest base number according to l.multiples
    """
    ytickorder = np.ceil(np.log10(span/ticks))
    minimalMultiplesy = np.argmin(np.abs(np.power(10,-ytickorder)*span/ticks-l.multiples))
    return l.multiples[minimalMultiplesy]*np.power(10,ytickorder)
    


def updateAxisDecorator(ax,direction='both'):
    def axisDecorator(func):
        @functools.wraps(func)
        def newFunc(*args,**kwargs):
            returnval = func(*args,**kwargs)
            ax.axisChanged(direction=direction)
            return returnval
        return newFunc
    return axisDecorator

def updateXAxisDecorator(ax):
    def axisDecorator(func):
        @functools.wraps(func)
        def newFunc(*args,**kwargs):
            returnval = func(*args,**kwargs)
            ax.xAxisChanged()
            return returnval
        return newFunc
    return axisDecorator

class MaxNLocator(mticker.MaxNLocator):
    def __init__(self, nbins=10, steps=None,
                trim=True,
                integer=False,
                symmetric=False,
                prune=None):
        # trim argument has no effect. It has been left for API compatibility
        super(MaxNLocator,self).__init__(nbins, steps=steps,
                                    integer=integer,
                                    symmetric=symmetric, prune=prune)
        self.create_dummy_axis()
        self._factor = 1#0.0#None

    def __call__(self, v1, v2): # pragma: no cover
        if self._factor is not None:
            self.set_bounds(v1*self._factor, v2*self._factor)
            locs = mticker.MaxNLocator.__call__(self)
            return np.array(locs), len(locs), self._factor
        else:
            self.set_bounds(v1, v2)
            locs = mticker.MaxNLocator.__call__(self)
            return np.array(locs), len(locs)

    def set_factor(self, f):
        self._factor = f


class MultipleLocator(mticker.MultipleLocator):
    def __init__(self,base=None):
        if base is None:
            base = 0.25
        super(MultipleLocator,self).__init__(base)
        self.create_dummy_axis()
        self._factor = 1#0.0
        self._multiplerVals = np.array([1,2,4,5,10])
        self.multiples = 1.0/self.multiplerVals

    @property
    def multiplerVals(self):
        return self._multiplerVals

    @multiplerVals.getter
    def multiplerVals(self):
        return self._multiplerVals

    @multiplerVals.setter
    def multiplerVals(self,multiplerVals):
        self._multiplerVals = multiplerVals
        self.multiples = 1.0/multiplerVals


    def __call__(self, v1, v2): # pragma: no cover
        if self._factor is not None:
            self.axis.set_view_interval(v1*self._factor, v2*self._factor)
            locs = mticker.MultipleLocator.__call__(self)
            

            return np.array(locs), len(locs), self._factor
        else:
            self.axis.set_view_interval(v1, v2)
            locs = mticker.MultipleLocator.__call__(self)
            return np.array(locs), len(locs), None

    def set_factor(self, f):
        self._factor = f

def forceGridUpdate(self):
    self._grid_helper._force_update = True
    self.pchanged()
    self.stale = True

def get_aspect(ax):
    figW, figH = ax.get_figure().get_size_inches()
    _, _, w, h = ax.get_position().bounds
    disp_ratio = (figH * h) / (figW * w)
    data_ratio = sub(*ax.get_ylim()) / sub(*ax.get_xlim())
    return (disp_ratio / data_ratio)

def axisChanged(axis,forceUpdate=False,direction='both'):
    """Function to recalculate the base number for RLU axis"""
    s = axis.sample
    xlim = axis.get_xlim()
    ylim = axis.get_ylim()
    xlimDiff = np.diff(xlim)
    ylimDiff = np.diff(ylim)
    
    if direction == 'both':
        direction = ['x','y']
        difference = [xlimDiff,ylimDiff]
        locators = [axis._grid_helper.grid_finder.grid_locator1,axis._grid_helper.grid_finder.grid_locator2]
    else:
        if direction == 'x':
            difference = [xlimDiff]  
            locators = [axis._grid_helper.grid_finder.grid_locator1]
        else:
            difference = [ylimDiff]  
            locators = [axis._grid_helper.grid_finder.grid_locator2]
        direction = [direction]
    for d,diff,locator in zip(direction,difference,locators):
        forceUpdate = True
        if isinstance(locator,MultipleLocator):
            if not np.isclose(getattr(axis,'_old{}limDiff'.format(d.capitalize())),diff) \
            or forceUpdate:# if new image size

                points = np.array(np.meshgrid(xlim,ylim)).reshape(-1,2) # Generate the 4 corner points
                Qs = np.array([s.inv_tr(p[0],p[1]) for p in points])
                if d=='x':
                    span = np.max(Qs[:,0])-np.min(Qs[:,0])
                else:
                    span = np.max(Qs[:,1])-np.min(Qs[:,1])
                if not hasattr(axis,'{}ticks'.format(d)):
                    setattr(axis,'{}ticks'.format(d),7)
                
                ticks = getattr(axis,'{}ticks'.format(d))
                setattr(axis,'_old{}limDiff'.format(d.capitalize()),diff)
            

            if not hasattr(axis,'{}base'.format(d)):
                base = calculateBase(locator,span,ticks)
            else:
                base = getattr(axis,'{}base'.format(d))

            locator.set_params(base=base)
        
        elif isinstance(locator,MaxNLocator):
            if hasattr(axis,'{}ticks'.format(d)):
                ticks = getattr(axis,'{}ticks'.format(d))
            else:
                ticks = 7
            locator.set_params(nbins = ticks)
        else:
            return 
    
    # force an update
    axis.forceGridUpdate()



def createRLUAxes(self,figure=None,ids=[1, 1, 1],basex=None,basey=None,projection=2,step=0):
    """Create a reciprocal lattice plot for a given DataSet object.
    
    Args:
        
        - Dataset (DataSet): DataSet object for which the RLU plot is to be made.

    Kwargs:

        - figure: Matplotlib figure in which the axis is to be put (default None)

        - ids (array): List of integer numbers provided to the SubplotHost ids attribute (default [1,1,1])

        - basex (float): Ticks are positioned at multiples of this value along x (default None)

        - basey (float): Ticks are positioned at multiples of this value along y (default None)

        - projection (int): Projection direction (default 2)

        - step (float): Projection along out-of-plane projection vector (default 0)

    Returns:
        
        - ax (Matplotlib axes): Axes containing the RLU plot.

    .. note::
        When rlu axis is created, the orientation of Qx and Qy is assumed to be rotated as well. 
        This is to be done in the self.View3D method call!

    .. note::
        When using python 2 the changing of tick marks is not supported due to limitations in matplotlib. However, if python 3 is used, the number 
        of ticks and their location can be change after the initialization using the set_xticks_number, set_yticks_number chaning the wanted number 
        of tick marks, or the set_xticks_base or set_yticks_base to change the base number, see RLU tutorial under Tools. As default a sufficient base
        number is found and will update when zooming.
        
    """

    sample = copy.deepcopy(self[0].sample)

    if figure is None:
        fig = plt.figure(figsize=(7, 4))
    else:
        fig = figure
    def calculateTicks(ticks,angle,round=True):
        val = ticks/np.tan(angle/2.0)
        if round:
            return np.array(np.round(val),dtype=int)
        else:
            return val


    if  not basex is None or not basey is None: # Either basex or basey is provided (or both)
        if basex is None:
            basex = calculateTicks(1.0,lambda: sample.projectionAngle(projection=projection),round=False)
        elif basey is None:
            basey = basex/calculateTicks(1.0,lambda: sample.projectionAngle(projection=projection),round=False)

        grid_locator1 = MultipleLocator(base=basex)
        grid_locator2 = MultipleLocator(base=basey)
    else:
        basex = 0.5
        basey = 0.5

        grid_locator1 = MultipleLocator(base=basex)
        grid_locator2 = MultipleLocator(base=basey)
        
    grid_helper = GridHelperCurveLinear((lambda x,y: sample.tr(x,y,projection=projection), lambda x,y: sample.inv_tr(x,y,projection=projection))
                                        ,grid_locator1=grid_locator1,grid_locator2=grid_locator2)

    ax = SubplotHost(fig, *ids, grid_helper=grid_helper)
    ax.sample = sample
    ax._step = step

    
    ax.basex = basex
    ax.basey = basey

    def set_axis(ax,v1,v2,*args):
        if args != ():
            points = np.concatenate([[v1,v2],[x for x in args]],axis=0)
        else:
            points = np.array([v1,v2])
            
        if points.shape[1] == 3:
            points = ax.sample.calculateHKLtoProjection(points[:,0],points[:,1],points[:,2]).T
        boundaries = np.array([ax.sample.inv_tr(x[0],x[1]) for x in points])
        ax.set_xlim(boundaries[:,0].min(),boundaries[:,0].max())
        ax.set_ylim(boundaries[:,1].min(),boundaries[:,1].max())
        ax.forceGridUpdate()


    fig.add_subplot(ax)
    ax.set_aspect(1.)
    ax.grid(True, zorder=0)

    if not np.isclose(ax.sample.projectionAngle(projection=projection),np.pi/2.0,atol=0.001):
        ax.axis["top"].major_ticklabels.set_visible(True)
        ax.axis["right"].major_ticklabels.set_visible(True)

    def advancedFormat_coord(ax,x,y,projection):
        z = np.ones_like(x)*ax._step
        return ax.sample.format_coord(x,y,z,projection=projection)

    ax.format_coord = lambda x,y: advancedFormat_coord(ax,x,y,projection=projection)
    ax.set_axis = lambda v1,v2,*args: set_axis(ax,v1,v2,*args)

    def beautifyLabel(vec):
        Vec = [x.astype(int) if np.isclose(x.astype(float)-x.astype(int),0.0) else x.astype(float) for x in vec]
        return '{} [RLU]'.format(', '.join([str(x) for x in Vec]))


    projectionMat = np.delete(ax.sample.projectionVectors,projection,axis=1)
    
    ax.set_xlabel(beautifyLabel(projectionMat[:,0]))
    ax.set_ylabel(beautifyLabel(projectionMat[:,1]))



    ax.calculateTicks = lambda value:calculateTicks(value,ax.sample.projectionAngle)
    ax.forceGridUpdate = lambda:forceGridUpdate(ax)
    ax._oldXlimDiff = np.diff(ax.get_xlim())
    ax._oldYlimDiff = np.diff(ax.get_ylim())

    ax.get_aspect_ratio = lambda: get_aspect(ax)

    ax.callbacks.connect('xlim_changed', axisChanged)
    ax.callbacks.connect('ylim_changed', axisChanged)
    ax.callbacks.connect('draw_event',lambda ax: axisChanged(ax,forceUpdate=True))
    ax.axisChanged = lambda direction='both': axisChanged(ax,forceUpdate=True,direction=direction)

    @updateAxisDecorator(ax=ax,direction='x')
    def set_xticks_base(xBase,ax=ax):
        """Setter of the base x ticks to be used for plotting

        Args:

            - xBase (float): Base of the tick marks

        """
        if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MultipleLocator):
            l1 = MultipleLocator(base=xBase)
            ax._grid_helper.update_grid_finder(grid_locator1=l1)

        ax.xbase = xBase

    @updateAxisDecorator(ax=ax,direction='y')
    def set_yticks_base(yBase,ax=ax):
        """Setter of the base y ticks to be used for plotting

        Args:

            - yBase (float): Base of the tick marks

        """
        if not isinstance(ax._grid_helper.grid_finder.grid_locator2,MultipleLocator):
            l2 = MultipleLocator(base=yBase)
            ax._grid_helper.update_grid_finder(grid_locator2=l2)
        ax.ybase = yBase

    @updateAxisDecorator(ax=ax,direction='x')
    def set_xticks_number(xNumber,ax=ax):
        """Setter of the number of x ticks to be used for plotting

        Args:

            - xNumber (int): Number of x tick marks

        """
        if not isinstance(ax._grid_helper.grid_finder.grid_locator1,MaxNLocator):
            l1 = MaxNLocator(nbins=xNumber)
            ax._grid_helper.update_grid_finder(grid_locator1=l1)
        ax.xticks = xNumber

    @updateAxisDecorator(ax=ax,direction='y')
    def set_yticks_number(yNumber,ax=ax):
        """Setter of the number of y ticks to be used for plotting

        Args:

            - yNumber (int): Number of y tick marks

        """
        if not isinstance(ax._grid_helper.grid_finder.grid_locator2,MaxNLocator):
            l2 = MaxNLocator(nbins=yNumber)
            ax._grid_helper.update_grid_finder(grid_locator2=l2)
        ax.yticks = yNumber

    ax.set_xticks_base = set_xticks_base
    ax.set_yticks_base = set_yticks_base
    ax.set_xticks_number = set_xticks_number
    ax.set_yticks_number = set_yticks_number


    return ax





import h5py as hdf
import numpy as np
import pickle as pickle
import matplotlib.pyplot as plt
import pandas as pd
import os
from DMCpy import DataFile, _tools, Viewer3D, RLUAxes, TasUBlibDEG


class DataSet(object):
    def __init__(self, dataFiles=None,**kwargs):
        """DataSet object to hold a series of DataFile objects

        Kwargs:

            - dataFiles (list): List of data files to be used in reduction (default None)

        Raises:

            - NotImplementedError

            - AttributeError

        """

        if dataFiles is None:
            self.dataFiles = []
        else:
            if isinstance(dataFiles,(str,DataFile.DataFile)): # If either string or DataFile instance wrap in a list
                dataFiles = [dataFiles]
            try:
                self.dataFiles = [DataFile.loadDataFile(dF) if isinstance(dF,(str)) else dF for dF in dataFiles]
            except TypeError:
                raise AttributeError('Provided dataFiles attribute is not iterable, filepath, or of type DataFile. Got {}'.format(dataFiles))
            
            self._getData()

    def _getData(self):
        # Collect parameters listed below across data files into self
        for parameter in ['counts','monitor','twoTheta','correctedTwoTheta','fileName','pixelPosition','wavelength','mask','normalization','normalizationFile','time','temperature']:
            setattr(self,parameter,np.array([getattr(d,parameter) for d in self]))

        
        types = [df.fileType for df in self]
        if len(types)>1:
            if not np.all([types[0] == t for t in types[1:]]):
                raise AttributeError('Provided data files have different types!\n'+'\n'.join([df.fileName+': '+df.scanType for df in self]))
        self.type = types[0]


    def __len__(self):
        """return number of DataFiles in self"""
        return len(self.dataFiles)
        

    def __eq__(self,other):
        """Check equality to another object. If they are of the same class (DataSet) and have the same attribute keys, the compare equal"""
        return np.logical_and(set(self.__dict__.keys()) == set(other.__dict__.keys()),self.__class__ == other.__class__)


    def __getitem__(self,index):
        try:
            return self.dataFiles[index]
        except IndexError:
            raise IndexError('Provided index {} is out of bounds for DataSet with length {}.'.format(index,len(self)))

    def __len__(self):
        return len(self.dataFiles)
    
    def __iter__(self):
        self._index=0
        return self
    
    def __next__(self):
        if self._index >= len(self):
            raise StopIteration
        result = self.dataFiles[self._index]
        self._index += 1
        return result

    def next(self):
        return self.__next__()

    def append(self,item):
        try:
            if isinstance(item,(str,DataFile.DataFile)): # A file path or DataFile has been provided
                item = [item]
            for f in item:
                if isinstance(f,str):
                    f = DataFile.loadDataFile(f)
                self.dataFiles.append(f)
        except Exception as e:
            raise(e)
        self._getData()

    def __delitem__(self,index):
        if index < len(self.dataFiles):
            del self.dataFiles[index]
        else:
            raise IndexError('Provided index {} is out of bounds for DataSet with length {}.'.format(index,len(self.dataFiles)))
        self._getData


    @property
    def sample(self):
        return [df.sample for df in self]

    @sample.getter
    def sample(self):
        return [df.sample for df in self]

    @sample.setter
    def sample(self,sample):
        for df in self:
            df.sample = sample

    def generateMask(self,maskingFunction = DataFile.maskFunction, **pars):
        """Generate mask to applied to data in data file
        
        Kwargs:

            - maskingFunction (function): Function called on self.phi to generate mask (default maskFunction)

        All other arguments are passed to the masking function.

        """
        for d in self:
            d.generateMask(maskingFunction,**pars)
        self._getData()

    @_tools.KwargChecker()
    def sumDetector(self,twoThetaBins=None,applyNormalization=True,correctedTwoTheta=True,dTheta=0.125):
        """Find intensity as function of either twoTheta or correctedTwoTheta

        Kwargs:

            - twoThetaBins (list): Bins into which 2theta is to be binned (default min(2theta),max(2theta) in steps of 0.5)

            - applyNormalization (bool): If true, take detector efficiency into account (default True)

            - correctedTwoTheta (bool): If true, use corrected two theta, otherwise sum vertically on detector (default True)

        Returns:

            - twoTheta
            
            - Normalized Intensity
            
            - Normalized Intensity Error

            - Total Monitor

        """

        if correctedTwoTheta:
            twoTheta = self.correctedTwoTheta
        else:
            if len(self.twoTheta.shape) == 3: # shape is (df,z,twoTheta), needs to be passed as (df,n,z,twoTheta)
                twoTheta = self.twoTheta[:,np.newaxis].repeat(self.counts.shape[1],axis=1) # n = scan steps
            else:
                twoTheta = self.twoTheta

        if twoThetaBins is None:
            anglesMin = np.min(twoTheta)
            anglesMax = np.max(twoTheta)
            twoThetaBins = np.arange(anglesMin-0.5*dTheta,anglesMax+0.51*dTheta,dTheta)

        if self.type.lower() == 'singlecrystal':
            monitorRepeated = np.array([np.ones_like(df.counts)*df.monitor.reshape(-1,1,1) for df in self])
        else:
            monitorRepeated = np.repeat(np.repeat(self.monitor[:,np.newaxis,np.newaxis],self.counts.shape[-2],axis=1),self.counts.shape[-1],axis=2)
            monitorRepeated.shape = self.counts.shape

        summedRawIntensity, _ = np.histogram(twoTheta[np.logical_not(self.mask)],bins=twoThetaBins,weights=self.counts[np.logical_not(self.mask)])

        if applyNormalization:
            summedMonitor, _ = np.histogram(twoTheta[np.logical_not(self.mask)],bins=twoThetaBins,weights=monitorRepeated[np.logical_not(self.mask)]*self.normalization[np.logical_not(self.mask)])
        else:
            summedMonitor, _ = np.histogram(twoTheta[np.logical_not(self.mask)],bins=twoThetaBins,weights=monitorRepeated[np.logical_not(self.mask)])

        inserted, _  = np.histogram(twoTheta[np.logical_not(self.mask)],bins=twoThetaBins)
        
        normalizedIntensity = summedRawIntensity/summedMonitor
        normalizedIntensityError =  np.sqrt(summedRawIntensity)/summedMonitor

        return twoThetaBins, normalizedIntensity, normalizedIntensityError,summedMonitor
    

    @_tools.KwargChecker(function=plt.errorbar,include=_tools.MPLKwargs)
    def plotTwoTheta(self,ax=None,twoThetaBins=None,applyNormalization=True,correctedTwoTheta=True,dTheta=0.125,**kwargs):
        """Plot intensity as function of correctedTwoTheta or twoTheta

        Kwargs:

            - ax (axis): Matplotlib axis into which data is plotted (default None - generates new)

            - twoThetaBins (list): Bins into which 2theta is to be binned (default min(2theta),max(2theta) in steps of 0.1)

            - applyNormalization (bool): If true, take detector efficiency into account (default True)

            - correctedTwoTheta (bool): If true, use corrected two theta, otherwise sum vertically on detector (default True)

            - All other key word arguments are passed on to plotting routine

        Returns:

            - ax: Matplotlib axis into which data was plotted

            - twoThetaBins
            
            - normalizedIntensity
            
            - normalizedIntensityError

            - summedMonitor

        """
        
        
        twoThetaBins, normalizedIntensity, normalizedIntensityError,summedMonitor = self.sumDetector(twoThetaBins=twoThetaBins,applyNormalization=applyNormalization,\
                                                                                       correctedTwoTheta=correctedTwoTheta,dTheta=dTheta)

        TwoThetaPositions = 0.5*(twoThetaBins[:-1]+twoThetaBins[1:])

        if not 'fmt' in kwargs:
            kwargs['fmt'] = '-'

        if ax is None:
            fig,ax = plt.subplots()

        ax._errorbar = ax.errorbar(TwoThetaPositions,normalizedIntensity,yerr=normalizedIntensityError,**kwargs)
        ax.set_xlabel(r'$2\theta$ [deg]')
        ax.set_ylabel(r'Intensity [arb]')

        def format_coord(ax,xdata,ydata):
            if not hasattr(ax,'xfmt'):
                ax.mean_x_power = _tools.roundPower(np.mean(np.diff(ax._errorbar.get_children()[0].get_data()[0])))
                ax.xfmt = r'$2\theta$ = {:3.'+str(ax.mean_x_power)+'f} Deg'
            if not hasattr(ax,'yfmt'):
                ymin,ymax,ystep = [f(ax._errorbar.get_children()[0].get_data()[1]) for f in [np.min,np.max,len]]
                
                ax.mean_y_power = _tools.roundPower((ymax-ymin)/ystep)
                ax.yfmt = r'Int = {:.'+str(ax.mean_y_power)+'f} cts'

            return ', '.join([ax.xfmt.format(xdata),ax.yfmt.format(ydata)])

        ax.format_coord = lambda format_xdata,format_ydata:format_coord(ax,format_xdata,format_ydata)

        return ax,twoThetaBins, normalizedIntensity, normalizedIntensityError,summedMonitor

    def plotInteractive(self,ax=None,masking=True,**kwargs):
        """Generate an interactive plot of data.

        Kwargs:

            - ax (axis): Matplotlib axis into which the plot is to be performed (default None -> new)

            - masking (bool): If true, the current mask in self.mask is applied (default True)

            - Kwargs: Passed on to errorbar or imshow depending on data dimensionality

        Returns:

            - ax: Interactive matplotlib axis

        """
        if ax is None:
            fig,ax = plt.subplots()
        else:
            fig = ax.get_figure()
        
        twoTheta = self.twoTheta

        if self.type.lower() in ['singlecrystal','powder']:
            shape = self.counts.shape
            
            intensityMatrix = np.divide(self.counts,self.normalization*self.monitor[:,:,np.newaxis,np.newaxis]).reshape(-1,shape[2],shape[3])
            mask = self.mask.reshape(-1,shape[2],shape[3])
            ax.titles = np.concatenate([[df.fileName]*len(df.A3) for df in self],axis=0)
        else:
            # Find intensity
            intensityMatrix = np.divide(self.counts,self.normalization*self.monitor[:,np.newaxis,np.newaxis])
            mask = self.mask
            

        if masking is True: # If masking, apply self.mask
            intensityMatrix[mask] = np.nan

        # Find plotting limits (For 2D pixel limits found later)
        thetaLimits = [f(twoTheta) for f in [np.min,np.max]]
        intLimits = [f(intensityMatrix) for f in [np.nanmin,np.nanmax]]

        # Copy relevant data to the axis
        ax.intensityMatrix = intensityMatrix
        ax.intLimits = intLimits
        ax.twoTheta = twoTheta
        ax.twoThetaLimits = thetaLimits
        

        if not hasattr(kwargs,'fmt'):
            kwargs['fmt']='-'

        if self.type.upper() == 'OLD DATA': # Data is 1D, plot using errorbar
            ax.titles = [df.fileName for df in self]
            # calculate errorbars
            if 'colorbar' in kwargs: # Cannot be used for 1D plotting....
                del kwargs['colorbar']
            ax.errorbarMatrix = np.divide(np.sqrt(self.counts),self.normalization*self.monitor[:,np.newaxis,np.newaxis])
            def plotSpectrum(ax,index=0,kwargs=kwargs):
                if kwargs is None:
                    kwargs = {}
                if hasattr(ax,'_errorbar'): # am errorbar has already been plotted, delete ot
                    ax._errorbar.remove()
                    del ax._errorbar
                
                if hasattr(ax,'color'): # use the color from previous plot
                    kwargs['color']=ax.color
                
                if hasattr(ax,'fmt'):
                    kwargs['fmt']=ax.fmt

                # Plot data
                ax._errorbar = ax.errorbar(ax.twoTheta[index],ax.intensityMatrix[index],yerr=ax.errorbarMatrix[index].flatten(),**kwargs)
                ax.fmt = kwargs['fmt']
                ax.index = index # Update index and color
                ax.color = ax._errorbar.lines[0].get_color()
                # Set plotting limits and title
                ax.set_xlim(*ax.twoThetaLimits)
                ax.set_ylim(*ax.intLimits)
                ax.set_title(ax.titles[index])
                plt.draw()

                ax.set_ylabel('Inensity [arb]')
            
        elif self.type.upper() == 'POWDER':
            ax.titles = [df.fileName for df in self]
            # Find limits for y direction
            
            ax.twoTheta = np.array([df.twoTheta for df in self])
            ax.idxSpans = np.cumsum([len(df.A3) for df in self]) # limits of indices corresponding to data file limits
            ax.IDX = -1 # index of current data file
            ax.twoThetaLimits = [f(ax.twoTheta) for f in [np.nanmin,np.nanmax]]
            ax.pixelLimits = [-0.1,0.1]

            def plotSpectrum(ax,index=0,kwargs=kwargs):
                # find color bar limits
                vmin,vmax = ax.intLimits

                newIDX = np.sum(index>=ax.idxSpans)
                if newIDX != ax.IDX:
                    ax.IDX = newIDX
                    if hasattr(ax,'_pcolormesh'):
                        ax.cla()
                    ax._pcolormesh = ax.pcolormesh(self.twoTheta[ax.IDX],self.pixelPosition[ax.IDX,2],ax.intensityMatrix[index],shading='auto',vmin=vmin,vmax=vmax)
                
                elif hasattr(ax,'_pcolormesh'):
                    ax._pcolormesh.set_array(ax.intensityMatrix[index])
                else:
                    ax._pcolormesh = ax.pcolormesh(self.twoTheta[ax.IDX],self.pixelPosition[ax.IDX,2],ax.intensityMatrix[index],shading='auto',vmin=vmin,vmax=vmax)

                ax.index = index
                if 'colorbar' in kwargs: # If colorbar attribute is given, use it
                    if kwargs['colorbar']: 
                        if not hasattr(ax,'_colorbar'): # If no colorbar is present, create one
                            ax._colorbar = fig.colorbar(ax._pcolormesh,ax=ax)
                # Set limits
                ax.set_xlim(*ax.twoThetaLimits)
                ax.set_ylim(*ax.pixelLimits)
                ax.set_title(ax.titles[index])
                ax.set_aspect('auto')
                
                plt.draw()
                
                ax.set_ylabel('Intensity [arb]')

        elif self.type.lower() == 'singlecrystal':
            
            
            ax.A3 = np.concatenate([df.A3 for df in self],axis=0)
            ax.twoTheta = np.array([df.twoTheta for df in self])
            ax.idxSpans = np.cumsum([len(df.A3) for df in self]) # limits of indices corresponding to data file limits
            ax.IDX = -1 # index of current data file
            ax.twoThetaLimits = [f(ax.twoTheta) for f in [np.nanmin,np.nanmax]]
            ax.pixelLimits = [-0.1,0.1]

            def plotSpectrum(ax,index=0,kwargs=kwargs):
                # find color bar limits
                vmin,vmax = ax.intLimits

                newIDX = np.sum(index>=ax.idxSpans)
                if newIDX != ax.IDX:
                    ax.IDX = newIDX
                    if hasattr(ax,'_pcolormesh'):
                        ax.cla()
                    ax._pcolormesh = ax.pcolormesh(self.twoTheta[ax.IDX],self.pixelPosition[ax.IDX,2],ax.intensityMatrix[index],shading='auto',vmin=vmin,vmax=vmax)
                
                elif hasattr(ax,'_pcolormesh'):
                    ax._pcolormesh.set_array(ax.intensityMatrix[index])
                else:
                    ax._pcolormesh = ax.pcolormesh(self.twoTheta[ax.IDX],self.pixelPosition[ax.IDX,2],ax.intensityMatrix[index],shading='auto',vmin=vmin,vmax=vmax)

                
                ax.index = index
                if 'colorbar' in kwargs: # If colorbar attribute is given, use it
                    if kwargs['colorbar']: 
                        if not hasattr(ax,'_colorbar'): # If no colorbar is present, create one
                            ax._colorbar = fig.colorbar(ax._imshow,ax=ax)
                # Set limits
                ax.set_xlim(*ax.twoThetaLimits)
                ax.set_ylim(*ax.pixelLimits)
                #print(index)
                ax.set_title(ax.titles[index]+' - A3: {:.2f} [deg]'.format(ax.A3[index]))
                ax.set_aspect('auto')
                
                
                plt.draw()

            
            ax.set_ylabel(r'Pixel z position [m]')

        # For all cases, x axis is two theta in degrees
        ax.set_xlabel(r'2$\theta$ [deg]')
        # Add function as method
        ax.plotSpectrum = lambda index,**kwargs: plotSpectrum(ax,index,**kwargs)
        
        # Plot first data point
        ax.plotSpectrum(0)

        ##### Interactivity #####

        def increaseAxis(self,step=1): # Call function to increase index
            index = self.index
            index+=step
            if index>=len(self.intensityMatrix):
                index = len(self.intensityMatrix)-1
            self.plotSpectrum(index)
            
        def decreaseAxis(self,step=1): # Call function to decrease index
            index = self.index
            index-=step
            if index<=-1:
                index = 0
            self.plotSpectrum(index)

        # Connect functions to key presses
        def onKeyPress(self,event): # pragma: no cover
            if event.key in ['+','up']:
                increaseAxis(self)
            elif event.key in ['-','down']:
                decreaseAxis(self)
            elif event.key in ['home']:
                index = 0
                self.plotSpectrum(index)
            elif event.key in ['end']:
                index = len(self.intensityMatrix)-1
                self.plotSpectrum(index)
            elif event.key in ['pageup']: # Pressing page up or page down performs steps of 10
                increaseAxis(self,step=10)
            elif event.key in ['pagedown']:
                decreaseAxis(self,step=10)

        # Call function for scrolling with mouse wheel
        def onScroll(self,event): # pragma: no cover
            if(event.button=='up'):
                increaseAxis(self)
            elif event.button=='down':
                decreaseAxis(self)
        # Connect function calls to slots
        fig.canvas.mpl_connect('key_press_event',lambda event: onKeyPress(ax,event) )
        fig.canvas.mpl_connect('scroll_event',lambda event: onScroll(ax,event) )
        
        return ax


    def plotOverview(self,**kwargs):
        """Quick plotting of data set with interactive plotter and summed intensity.

        Kwargs:

            - masking (bool): If true, the current mask in self.mask is applied (default True)

            - kwargs (dict): Kwargs to be used for interactive or plotTwoTheta plot

        returns:

            - Ax (list): List of two axis, first containing the interactive plot, second summed two theta


        Kwargs for plotInteractiveKwargs:
        
            - masking (bool): Use generated mask for dataset (default True)

        Kwargs for plotTwoThetaKwargs:

            - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.1 Deg)
            
            - applyNormalization (bool): Use normalization files (default True)
            
            - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
        """

        fig,Ax = plt.subplots(2,1,figsize=(11,9),sharex=True)

        Ax = Ax.flatten()


        if not 'fmt' in kwargs:
            kwargs['fmt']='_'

        if not 'masking' in kwargs:
            kwargs['masking']= True

        if not 'twoThetaBins' in kwargs:
            kwargs['twoThetaBins']= None

        if not 'applyNormalization' in kwargs:
            kwargs['applyNormalization']= True


        if not 'correctedTwoTheta' in kwargs:
            kwargs['correctedTwoTheta']= True

        if not 'colorbar' in kwargs:
            kwargs['colorbar']= False

        plotInteractiveKwargs = {}
        for key in ['masking','fmt','colorbar']:
            plotInteractiveKwargs[key] = kwargs[key]
        
        plotTwoThetaKwargs = {}
        for key in ['twoThetaBins','fmt','correctedTwoTheta','applyNormalization']:
            plotTwoThetaKwargs[key] = kwargs[key]

        ax2,*_= self.plotTwoTheta(ax=Ax[1],**plotTwoThetaKwargs)
        ax = self.plotInteractive(ax = Ax[0],**plotInteractiveKwargs)

        ax.set_xlabel('')
        ax2.set_title('Integrated Intensity')

        fig.tight_layout()
        return Ax

    def Viewer3D(self,dqx,dqy,dqz,rlu=True,axis=2, raw=False,  log=False, grid = True, outputFunction=print, 
                 cmap='viridis',smart=False):

        """Generate a 3D view of all data files in the DatSet.
        
        Args:
        
            - dqx (float): Bin size along first axis in 1/AA
        
            - dqy (float): Bin size along second axis in 1/AA
        
            - dqz (float): Bin size along third axis in 1/AA

        Kwargs:

            - rlu (bool): Plot using reciprocal lattice units (default False)

            - axis (int): Initial view direction for the viewer (default 2)

            - raw (bool): If True plot counts else plot normalized counts (default False)

            - log (bool): Plot intensity as logarithm of intensity (default False)

            - grid (bool): Plot a grid on the figure (default True)

            - outputFunction (function): Function called when clicking on the figure (default print)

            - cmap (str): Name of color map used for plot (default viridis)
        
        """

        if rlu:
            rluAxesQxQy = self.createRLUAxes(projection=2)#**kwargs)
            figure = rluAxesQxQy.get_figure()
            figure.delaxes(rluAxesQxQy)
            rluAxesQxQz = self.createRLUAxes(figure=figure,projection=1)
            figure.delaxes(rluAxesQxQz)
            rluAxesQyQz = self.createRLUAxes(figure=figure,projection=0)
            figure.delaxes(rluAxesQyQz)
            axes = [rluAxesQyQz,rluAxesQxQz,rluAxesQxQy]


        else:
            axes = None

        Data,bins = self.binData3D(dqx,dqy,dqz,rlu=rlu,raw=raw,smart=smart)

        return Viewer3D.Viewer3D(Data,bins,axis=axis, ax=axes, grid=grid, log=log, outputFunction=outputFunction, cmap=cmap)

    def binData3D(self,dqx,dqy,dqz,rlu=True,raw=False,smart=False):

        maximas = []
        minimas = []
        for df in self:
            if rlu:
                pos = np.einsum('ij,jk',df.sample.ROT,df.q.reshape(3,-1))
                maximas.append(np.max(pos,axis=1))
                minimas.append(np.min(pos,axis=1))
            else:
                maximas.append(np.max(df.q.reshape(3,-1),axis=1))
                minimas.append(np.min(df.q.reshape(3,-1),axis=1))

        maximas = np.max(maximas,axis=0)
        minimas = np.min(minimas,axis=0)
        extremePositions = np.array([minimas,maximas]).T
        bins = _tools.calculateBins(dqx,dqy,dqz,extremePositions)

        returndata = None
        for df in self:
            
            

            if rlu:
                pos = np.einsum('ij,j...',df.sample.ROT,df.q).transpose(0,3,1,2) # shape -> steps,3,128,1152

            else:
                pos = df.q.transpose(1,0,2,3)# shape -> steps,3,128,1152

            if not raw:
                data = df.intensity#[np.logical_not(df.mask)]/df.normalization[np.logical_not(df.mask)] # shape steps,128,1152
            else:
                data = df.counts#[np.logical_not(df.mask)] # shape steps,128,1152
            # if smart:
            #     for p,d,mon in zip(pos,data,df.monitor):
            #         localReturndata,_ = _tools.binData3D(dqx,dqy,dqz,pos=p.reshape(3,-1),data=d.flatten(),bins = bins)

            #         if returndata is None:
            #             returndata = localReturndata
            #             returndata[-1]*=mon
            #         else:
            #             returndata[-1]*=mon
            #             for data,newData in zip(returndata,localReturndata):
            #                 data+=newData
            if True:
                pos = pos.transpose(1,0,2,3)
                localReturndata,_ = _tools.binData3D(dqx,dqy,dqz,pos=pos.reshape(3,-1),data=data.flatten(),bins = bins)

                if returndata is None:
                    returndata = localReturndata
                    returndata[-1]*=df.monitor[0]
                else:
                    returndata[-1]*=df.monitor[0]
                    for data,newData in zip(returndata,localReturndata):
                        data+=newData
                    

        intensities = np.divide(returndata[0],returndata[1])
        NaNs = returndata[1]==0
        intensities[NaNs]=np.nan
        return intensities,bins

    @_tools.KwargChecker()
    @_tools.overWritingFunctionDecorator(RLUAxes.createRLUAxes)
    def createRLUAxes(*args,**kwargs): # pragma: no cover
        raise RuntimeError('This code is not meant to be run but rather is to be overwritten by decorator. Something is wrong!! Should run {}'.format(RLUAxes.createRLUAxes))

        
    def plotCut1D(self,P1,P2,rlu=True,stepSize=0.01,width=0.05,widthZ=0.05,raw=False,optimize=True,ax=None,**kwargs):
        """Cut and plot data from P1 to P2 in steps of stepSize [1/AA] width a cylindrical width [1/AA]

        Args:

            - P1 (list): Start position for cut in either (Qx,Qy,Qz) or (H,K,L)

            - P2 (list): End position for cut in either (Qx,Qy,Qz) or (H,K,L)

        Kwargs:

            - rlu (bool): If True, P1 and P2 are in HKL, otherwise in QxQyQz (default True)

            - stepSize (float): Size of bins along cut direction in units of [1/AA] (default 0.01)

            - width (float): Integration width orthogonal to cut in units of [1/AA] (default 0.02)

            - raw (bool): If True, do not normalize data (default False)

            - optimize (bool): If True, perform optimized cutting (default True)

            - ax (matplotlib.axes): If None, a new is created (default None)

            - kwargs: All other kwargs are provided to the scatter plot of the axis

        Returns:

            - Pos,Int,Ax
            

        """

        hkl,I = self.cut1D(P1=P1,P2=P2,rlu=rlu,stepSize=stepSize,width=width,widthZ=widthZ,raw=raw,optimize=optimize)
        if ax is None:
            ax  = generate1DAxis(P1,P2,rlu=rlu)


        X = ax.calculatePositionInv(*hkl)
        ax.scatter(X,I,**kwargs)

        ax.get_figure().tight_layout()
        return hkl,I,ax

    def cut1D(self,P1,P2,rlu=True,stepSize=0.01,width=0.05,widthZ=0.05,raw=False,optimize=True):
        """Cut data from P1 to P2 in steps of stepSize [1/AA] width a cylindrical width [1/AA]

        Args:

            - P1 (list): Start position for cut in either (Qx,Qy,Qz) or (H,K,L)

            - P2 (list): End position for cut in either (Qx,Qy,Qz) or (H,K,L)

        Kwargs:

            - rlu (bool): If True, P1 and P2 are in HKL, otherwise in QxQyQz (default True)

            - stepSize (float): Size of bins along cut direction in units of [1/AA] (default 0.01)

            - width (float): Integration width orthogonal to cut in units of [1/AA] (default 0.02)

            - raw (bool): If True, do not normalize data (default False)

            - optimize (bool): If True, perform optimized cutting (default True)

        Returns:

            - Pos,Int....
            

        """
        intensities = None
        for df in self:
            if rlu:
                QStart = df.sample.calculateHKLToQxQyQz(*P1)
                QStop  = df.sample.calculateHKLToQxQyQz(*P2)
            else:
                QStart = P1
                QStop  = P2
            
            
            directionVector = (QStop-QStart).reshape(3,1)
            length = np.linalg.norm(directionVector)
            if np.isclose(length,0.0):
                raise AttributeError('The vector connecting the cut points has length 0. Received P1={}, P2={}'.format(','.join([str(x) for x in P1]),','.join([str(x) for x in P2])))
            directionVector*=1.0/length
            
            stopAlong = np.dot(QStop-QStart,directionVector)[0]
            sign = np.sign(stopAlong)
            
            bins = np.arange(-stepSize*0.5,np.abs(stopAlong)+stepSize*0.51,stepSize)
            
            intensity = []
            pos = []
            
            
            if not raw:
                data = df.intensity
            else:
                data = df.counts
            if optimize:
               
                optimizationStepInPlane = 0.005
                optimizationStepInPlane = np.min([optimizationStepInPlane,width*0.6])
                
                ## Define boundig box
                direction = QStop-QStart
                directionLength=np.linalg.norm(direction)
                direction*=1.0/directionLength
                orthogonal = np.cross(direction,np.array([0,0,1]))
                
                # Factor between actual cut and width used for cutoff
                expansionFactior = 1.5
                effectiveWidth = expansionFactior*width
                    
                if not np.isclose(np.abs(np.dot(direction,[0,0,1])),1.0): # If cut is not along z
                    # Points for bounding box
                    startEdge = QStart.reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonal.reshape(3,1)-stepSize*direction.reshape(3,1)
                    endEdge = QStop.reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonal.reshape(3,1)+stepSize*direction.reshape(3,1)
                    rightEdge = QStart.reshape(3,1)+0.5*effectiveWidth*orthogonal.reshape(3,1)+np.arange(-stepSize,directionLength+stepSize,optimizationStepInPlane)*direction.reshape(3,1)
                    leftEdge =  QStart.reshape(3,1)-0.5*effectiveWidth*orthogonal.reshape(3,1)+np.arange(-stepSize,directionLength+stepSize,optimizationStepInPlane)*direction.reshape(3,1)
                    
                    checkPositions = np.concatenate([startEdge,endEdge,rightEdge,leftEdge],axis=1)
                
                else: # if cut is along z
                    orthogonalX = np.array([1,0,0])
                    orthogonalY = np.array([0,1,0])
                    startEdge = np.mean([QStart,QStop],axis=0).reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonalX.reshape(3,1)-effectiveWidth*orthogonalY.reshape(3,1)
                    endEdge = np.mean([QStart,QStop],axis=0).reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonalX.reshape(3,1)+effectiveWidth*orthogonalY.reshape(3,1)
                    rightEdge = np.mean([QStart,QStop],axis=0).reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonalY.reshape(3,1)-effectiveWidth*orthogonalX.reshape(3,1)
                    leftEdge =  np.mean([QStart,QStop],axis=0).reshape(3,1)+np.arange(-effectiveWidth*0.5,effectiveWidth*0.51,optimizationStepInPlane).reshape(1,-1)*orthogonalY.reshape(3,1)+effectiveWidth*orthogonalX.reshape(3,1)
                    
                    checkPositions = np.concatenate([startEdge,endEdge,rightEdge,leftEdge],axis=1)
                
                # Calcualte the corresponding A3 and A4 positons
                E = np.power(df.ki[1,0][0]/0.694692,2.0)
                A3,A4 = np.array([TasUBlibDEG.converterToA3A4(*pos,E,E) for pos in checkPositions.T]).T
                
                # remove nan-values
                A4NonNaN = np.logical_not(np.isnan(A4))
                A3 = A3[A4NonNaN]
                A4 = A4[A4NonNaN]

                
                A3Min,A3Max = [f(A3) for f in [np.nanmin,np.nanmax]]
                A4Min,A4Max = [f(A4) for f in [np.nanmin,np.nanmax]]
                
                
                # Find and sort ascending the indices        
                twoThetaIdx = np.sort(np.array([np.argmin(np.abs(df.twoTheta[0]-tt)) for tt in [A4Min,A4Max]]))
                A3Idx = np.sort(np.array([np.argmin(np.abs(df.A3-a3)) for a3 in [A3Min,A3Max]]))

                if not np.isclose(np.abs(np.dot(direction,[0,0,1])),1.0):
                    maxQz = np.max([QStart[2],QStop[2]])+widthZ*expansionFactior
                    minQz = np.min([QStart[2],QStop[2]])-widthZ*expansionFactior
                    qzIdx = np.array(np.sort(np.array([np.argmin(np.abs(w-df.q[2,0,:,0])) for w in [minQz,maxQz]])))
                else:
                    qzIdx = np.array([0,df.counts.shape[1]])#np.sort(np.array([np.argmin(np.abs(p[2]-df.q[2,0,:,0])) for p in [P1,P2]])))
                
                mask = np.zeros_like(df.counts,dtype=bool)
                mask[A3Idx[0]:A3Idx[1]+1,
                        qzIdx[0]:qzIdx[1]+1,
                        twoThetaIdx[0]:twoThetaIdx[1]+1]=True
                
                data = data[mask]
                relativePosition = df.q[:,mask]-QStart.reshape(3,-1)
                
            else:
                # along = np.einsum('ij,i...->...j',relativePosition,directionVector)
                
                # orthogonal = np.linalg.norm(relativePosition-along*directionVector,axis=0)
                
                # orthogonal = np.linalg.norm(relativePosition-along*directionVector,axis=0)
                # test1 = (orthogonal<width).flatten()
                # test2 = (along[0]>-stepSize).flatten()
                # test3 = (along[0]<np.linalg.norm(stopAlong)+stepSize).flatten()
                
                # insideQ = np.all([test1,test2,test3],axis=0)
            
                # intensity = data.flatten()[insideQ]
                # pos = sign*along.flatten()[insideQ]
                
            #else:
                relativePosition = df.q.reshape(3,-1)-QStart.reshape(3,-1)
            
            along = np.einsum('ij,i...->...j',relativePosition,directionVector)
                
            orthogonal = np.linalg.norm(relativePosition-along*directionVector,axis=0)
            
            orthogonal = np.linalg.norm(relativePosition-along*directionVector,axis=0)
            test1 = (orthogonal<width*0.5).flatten()
            test2 = (along[0]>-stepSize).flatten()
            test3 = (along[0]<np.linalg.norm(stopAlong)+stepSize).flatten()
            
            insideQ = np.all([test1,test2,test3],axis=0)
        
            intensity = data.flatten()[insideQ]
            pos = sign*along.flatten()[insideQ]

                
        
            if intensities is None:
                intensities = np.histogram(pos,bins=bins,weights=intensity)[0]
                normCounts = np.histogram(pos,bins=bins)[0]
                monitors = np.full_like(intensities,df.monitor[0])
            else:
                intensities+=np.histogram(pos,bins=bins,weights=intensity)[0]
                normCounts+=np.histogram(pos,bins=bins)[0]
                monitors += np.full_like(intensities,df.monitor[0])
        
        I = np.divide(intensities,monitors)
        I[normCounts==0]=np.nan
        binCentres = 0.5*(bins[:-1]+bins[1:])
        
        positionVector = directionVector*binCentres+QStart.reshape(3,1)
        if rlu:
            positionVector = np.array([self[-1].sample.calculateQxQyQzToHKL(*bC) for bC in positionVector.T]).T
        return positionVector,I


    def autoAlignScatteringPlane(self,scatteringNormal,threshold=30,dx=0.04,dy=0.04,dz=0.08,distanceThreshold=0.15):
        """Automatically align scattering plane and peaks within
        
        Args:
            
            - scatteringNormal (vector 3D): Normal to the scattering plane in HKL
            
            
        Kwargs:
            
            - threshold (float): Thresholding for intensities within the 3D binned data used in peak search (default 30)
            
            - dx (float): size of 3D binning along Qx (default 0.04)
            
            - dy (float): size of 3D binning along Qy (default 0.04)
            
            - dz (float): size of 3D binning along Qz (default 0.08)
            
            - distanceThreshold (float): Distance in 1/AA where peaks are clustered together (default 0.15)
          
            
        This methods is an attempt to automatically align the scattering plane of all data files
        within the DataSet. The algorithm works as follows for each data file individually :
            
            1) Perform a 3D binning of data in to equi-sized bins with size (dx,dy,dz)
            
            2) "Peaks" are defined as all positions having intensities>threshold
        
            3) These peaks are clustered together if closer than 0.02 1/AA and centre of gravity
               using intensity is applied to find common centre.
               
            4) Above step is repeated with custom distanceThreshold
        
            5) Plane normals are found as cross products between all vectors connecting 
               all found peaks -> Gives a list of approximately NPeaks*(NPeaks-1)*(NPeaks-2)
               
            6) Plane normals are clustered and most common is used
        
            7) Peaks are rotated into the scattering plane
        
            8) Within the plane all found peaks are projected along the 'nice' plane vectors and
               the peak having the scattering length closest to an integer multiple of either is
               chosen for alignment
               
            9) Rotation within the plane is found by rotating the found peak either along x or y 
               depending on which projection vector was closest
               
           10) Sample is updated with found rotations.
        
        
        On the sample object, the peak used for alignment within the scattering plane is
        saved as sample.peakUsedForAlignment as a dictionary with 'HKL' and 'QxQyQz' holding
        suggested HKL point and original QxQyQz position.
        
        
        """
        
        for df in self:
            
            # 1) 
            Intensities,bins = _tools.binData3D(dx,dy,dz,df.q.reshape(3,-1),df.intensity)
            Intensities = np.divide(Intensities[0],Intensities[1])
            
            
            # 2)
            possiblePeaks = Intensities>threshold
            ints = Intensities[possiblePeaks]
            
            centerPoints = [b[:-1,:-1,:-1]+0.5*dB for b,dB in zip(bins,[dx,dy,dz])]
            
            positions = np.array([b[possiblePeaks] for b in centerPoints]).T
            
            
            
            # 3) assuming worse resolution out of plane
            distanceFunctionLocal = lambda a,b: _tools.distance(a,b,dx=1.0,dy=1.0,dz=0.5)
            peaksInitial = _tools.clusterPoints(positions,ints,distanceThreshold=0.02,distanceFunction=distanceFunctionLocal) 
            
            peakPositions = [p.position for p in peaksInitial]
            peakWeights   = [p.weight for p in peaksInitial]
            
            # 4) 
            peaks = _tools.clusterPoints(peakPositions,peakWeights,distanceThreshold=distanceThreshold,distanceFunction=distanceFunctionLocal)
            
            
            foundPeakPositions = np.array([p.position for p in peaks])
            
            
            # 5) 
            tripletNormal = _tools.calculateTriplets(foundPeakPositions,normalized=True)
            
            
            # 6) Combine the normal vectors closest to each other.
            normalVectors = _tools.clusterPoints(tripletNormal,np.ones(len(tripletNormal)),distanceThreshold=0.01)
            
            # Find the most frequent normal vector as the one with highest weight
            bestNormalVector = normalVectors[np.argmax([p.weight for p in normalVectors])].position
            
            # 7) 
            # Find rotation matrix transforming bestNormalVector to lay along the z-axis
            # Rotation is performed around the vector perpendicular to bestNormalVector and z-axis
            rotationVector = np.cross([0,0,1.0],bestNormalVector)
            rotationVector*=1.0/np.linalg.norm(rotationVector)
            # Rotation angle is given by the regular cosine relation, but due to z being [0,0,1] and both normal
            
            theta = np.arccos(bestNormalVector[2]) # dot(bestNormalVector,[0,0,1])/(lengths) <-- both unit vectors
            
            RotationToScatteringPlane = _tools.rotMatrix(rotationVector, theta,deg=False)
                
            # Rotated all found peaks to the scattering plane
            rotatedPeaks = np.einsum('ji,...j->...i',RotationToScatteringPlane,foundPeakPositions)
            
            # 8) Start by finding in plane vectors using provided scattering plane normal
            scatteringNormal = _tools.LengthOrder(np.asarray(scatteringNormal,dtype=float)) # 
            
            # Calculate into Q
            scatteringNormalB = np.dot(df.sample.B,scatteringNormal)
            
            # Find two main "nice" vectors in the scattering plane by finding a vector orthogonal to scatteringNormalB
            InPlaneGuess = np.cross(scatteringNormalB,np.array([1,-1,0]))
            
            # If scatteringNormalB happens to be along [1,-1,0] we just try again!
            if np.isclose(np.linalg.norm(InPlaneGuess),0.0,atol=1e-3):
                InPlaneGuess = np.cross(scatteringNormalB,np.array([1,1,0]))
                
            # planeVector1 is ensured to be orthogonal to scatteringNormalB 
            # (rounding is needed to better beautify the vector)
            planeVector1 = np.round(np.cross(InPlaneGuess,scatteringNormalB),4)
            planeVector1 = _tools.LengthOrder(planeVector1)
            
            # Second vector is radially found orthogonal to scatteringNormalB and planeVector1
            planeVector2 = np.round(np.cross(scatteringNormalB,planeVector1),4)
            planeVector2 = _tools.LengthOrder(planeVector2)
            
            # Try to align the last free rotation within the scattering plane by calculating
            # the length of the scattering vectors for the peaks and compare to moduli of 
            # planeVector1 and planeVector2
            
            lengthPeaksInPlane = np.linalg.norm(rotatedPeaks,axis=1)
            
            planeVector1Length = np.linalg.norm(np.dot(df.sample.B,planeVector1))
            planeVector2Length = np.linalg.norm(np.dot(df.sample.B,planeVector2))
            
            projectionAlongPV1 = lengthPeaksInPlane/planeVector1Length
            projectionAlongPV2 = lengthPeaksInPlane/planeVector2Length
            
            # Initialize the along1 and along2 with False to force while loop
            along1 = [False]
            along2 = [False]
            
            atol = 0.001
            # While there are no peaks found along with a modulus close to 0/1 along the main directions 
            # iteratively increase tolerance (atol)
            while np.sum(along1)+np.sum(along2) == 0: 
                atol+=0.001
                along1 = np.array([np.logical_or(np.isclose(x,0.0,atol=atol),np.isclose(x,1.0,atol=atol)) for x in np.mod(projectionAlongPV1,1)])
                along2 = np.array([np.logical_or(np.isclose(x,0.0,atol=atol),np.isclose(x,1.0,atol=atol)) for x in np.mod(projectionAlongPV2,1)])
                
                
            
            # Either we found a peak along planeVector1 or planeVector2
            if np.sum(along1)> 0:
                foundPosition = rotatedPeaks[along1][0]
                axisOffset = 0.0 # We want planeVector1 to be along the x-axis
                peakUsedForAlignment = {'HKL':   planeVector1*projectionAlongPV1[along1][0],
                                        'QxQyQz':foundPeakPositions[along1][0]}
            
            else:
                foundPosition = rotatedPeaks[along2][0]
                axisOffset = 90.0 # planeVector2 is along the y-axis, e.i. 90 deg rotated from x
                peakUsedForAlignment = {'HKL':   planeVector2*projectionAlongPV2[along2][0],
                                        'QxQyQz':foundPeakPositions[along2][0]}
        
        
            
            # 9) 
            # Calculate the actual position of the peak found along planeVector1 or planeVector2
            offsetA3 = np.rad2deg(np.arctan2(foundPosition[1],foundPosition[0]))-axisOffset
            
            # Find rotation matrix which is around the z-axis and has angle of -offsetA3
            rotation = np.dot(_tools.rotMatrix(np.array([0,0,1.0]),-offsetA3),RotationToScatteringPlane.T)
            
            
            # 10) 
            # sample rotation has now been found (converts between instrument 
            # qx,qy,qz to qx along planeVector1 and qy along planeVector2)
            
            sample = df.sample
            sample.ROT = rotation
            sample.P1 = _tools.LengthOrder(planeVector1)
            sample.P2 = _tools.LengthOrder(planeVector2)
            sample.P3 = _tools.LengthOrder(scatteringNormal)
            
            sample.projectionVectors = np.array([sample.P1,sample.P2,sample.P3]).T
            
            sample.projectionB = np.diag(np.linalg.norm(np.dot(sample.projectionVectors.T,sample.B),axis=1))
            sample.UB = np.dot(sample.ROT.T,np.dot(sample.projectionB,np.linalg.inv(sample.projectionVectors)))
            
            sample.peakUsedForAlignment = peakUsedForAlignment
        

    def export_PSI_format(self,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,outFolder=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

        """
        The function takes a data set and merge the files.
        Outputs a .dat file in PSI format (Fullprof inst. 8)
        Saves the file with input name
        Data files used in the export is given in output file
        
        Kwargs:
            
            - dTheta (Float): stepsize of binning if no nins is given (default is 0.125)
            
            - twoThetaOffset (float): Linear shift of two theta, default is 0. To be used if a4 in hdf file is incorrect
            
            - Bins (list): Bins into which 2theta is to be binned (default min(2theta),max(2theta) in steps of 0.125)
            
            - outFile (str): String that will be used for outputfile. Default is automatic generated name.

            - outFolder (str): Path to folder data will be saved. Default is current working directory.
            
        - Arguments for automatic file name:
                
            - sampleName (bool): Include sample name in filename. Default is True.
        
            - temperature (bool): Include temperature in filename. Default is True.
        
            - magneticField (bool): Include magnetic field in filename. Default is False.
        
            - electricField (bool): Include electric field in filename. Default is False.
        
            - fileNumber (bool): Include sample number in filename. Default is True.
            
        Kwargs for sumDetector:

            - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
                
            - applyNormalization (bool): Use normalization files (default True)
                
            - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
            
        Returns:
            
            .dat file in PSI format with input name
            
        Note: Input is a data set.
            
        Example:
            >>> inputNumber = _tools.fileListGenerator(565,folder)
            >>> ds = DataSet.DataSet(inputNumber)
            >>> for df in ds:
            ...    if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
            ...        df.monitor = np.ones_like(df.monitor)
            >>> export_PSI_format(ds)
        
        """

        twoTheta = self.twoTheta
        
        anglesMin = np.min(twoTheta)
        anglesMax = np.max(twoTheta)
        
        if bins is None:
            bins = np.arange(anglesMin-0.5*dTheta,anglesMax+0.51*dTheta,dTheta)
        
        bins,intensity,err,monitor = self.sumDetector(bins,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta)
        
        bins = bins + twoThetaOffset
        
        # find mean monitor
        meanMonitor = np.median(monitor)
        intensity[np.isnan(intensity)] = -1
        
        # rescale intensity and err
        intensity*=meanMonitor
        err*=meanMonitor
        
        step = np.mean(np.diff(bins))
        start = bins[0]+0.5*step
        stop = bins[-1]-0.5*step
        
        temperatures = [df.temperature for df in self]
        meanTemp = np.mean(temperatures)
        stdTemp = np.std(temperatures)

        if np.all([x == self.sample[0].name for x in [s.name for s in self.sample[1:]]]):
            samName = self.sample[0].name        #.decode("utf-8")
        else:
            samName ='Unknown! Combined different sample names'
        
        if np.all([np.isclose(x,self.wavelength[0]) for x in self.wavelength[1:]]):
            wavelength = self.wavelength[0]
        else:
            wavelength ='Unknown! Combined different Wavelengths'
        

        # reshape intensity and err to fit into (10,x)
        intNum = len(intensity)
        
        # How many empty values to add to allow reshape
        addEmpty = int(10*np.ceil(intNum/10.0)-intNum)
        
        intensity = np.concatenate([intensity,addEmpty*[np.nan]]).reshape(-1,10)
        err = np.concatenate([err,addEmpty*[np.nan]]).reshape(-1,10)
        
        ## Generate output to DMC file format
        titleLine = "DMC, "+samName
        paramLine = "lambda={:9.5f}, T={:8.3f}, dT={:7.3f}, Date='{}'".format(wavelength,meanTemp,stdTemp,self[0].startTime)#.decode("utf-8"))
        paramLine2= ' '+' '.join(["{:7.3f}".format(x) for x in [start,step,stop]])+" {:7.0f}".format(meanMonitor)+'., sample="'+samName+'"'
        
        dataLinesInt = '\n'.join([' '+' '.join(["{:6.0f}.".format(x).replace('nan.','    ') for x in line]) for line in intensity])
        dataLinesErr = '\n'.join([' '+' '.join(["{:7.1f}".format(x).replace('nan.','    ') for x in line]) for line in err])
        
        ## Generate bottom information part
        if len(self) == 1:
            year = 2022
            fileNumbers = str(int(self.fileName[0].split('n')[-1].split('.')[0]))
        else:
            year,fileNumbers = _tools.numberStringGenerator(self.fileName)
        
        fileList = " Filelist='dmc:{}:{}'".format(year,fileNumbers)
        
        minmax = [np.nanmin,np.nanmax]
        
        twoThetaStart = self.twoTheta[:,0]
        twoTheta = [np.min(twoThetaStart),np.max(twoThetaStart)]
        Counts = [int(func(intensity)) for func in minmax]
        numor = fileNumbers.replace('-',' ')
        Npkt = len(bins) - 1        
        
        owner = self[-1].user#.decode("utf-8")
        a1 = self[-1].monochromatorRotationAngle[0]
        a2 = self[-1].monochromatorTakeoffAngle[0]
        a3 = self[-1].A3[0]
        mcv = self[-1].monochromatorCurvature[0]
        mtx = self[-1].monochromatorTranslationLower[0]
        mty = self[-1].monochromatorTranslationUpper[0]
        mgu = self[-1].monochromatorGoniometerUpper[0]
        mgl = self[-1].monochromatorGoniometerLower[0]
        
        bMon = [df.protonBeam for df in self]
        pMon = [df.monitor for df in self]
        sMon = [[0.0]]
        
        timeMin, timeMax = [func(self.time) for func in minmax]
        sMonMin, sMonMax = [func(sMon) for func in minmax]
        bMonMin, bMonMax = [func(bMon) for func in minmax]
        aMon = np.mean([0.0 for df in self])
        pMonMin, pMonMax = [func(pMon) for func in minmax]
        muR = 0.0                           #self[-1].sample.sample_mur[0]
        preset = self[-1].mode      #.decode("utf-8")
        
        paramLines = []
        paramLines.append(" a4={:1.1f}. {:1.1f}.; Counts={} {}; Numor={}; Npkt={}; owner='{}'".format(*twoTheta,*Counts,numor,Npkt,owner))
        paramLines.append('  a1={:4.2f}; a2={:3.2f}; a3={:3.2f}; mcv={:3.2f}; mtx={:3.2f}; mty={:3.2f}; mgu={:4.3f}; mgl={:4.3f}; '.format(a1,a2,a3,mcv,mtx,mty,mgu,mgl))
        paramLines.append('  time={:4.4f} {:4.4f}; sMon={:4.0f}. {:4.0f}.; bMon={:3.0f}. {:3.0f}.; aMon={:1.0f}'.format(timeMin,timeMax,sMonMin,sMonMax,bMonMin,bMonMax,aMon))
        paramLines.append("  pMon={:7.0f}. {:7.0f}.; muR={:1.0f}.; Preset='{}'".format(float(pMonMin),float(pMonMax),muR,preset))
        paramLines.append("  calibration='{}'".format(self[-1].normalizationFile))
        paramLines.append("")
        fileString = '\n'.join([titleLine,paramLine,paramLine2,dataLinesInt,dataLinesErr,fileList,*paramLines])
        
        # get magnetic field
        # get electric field
        mag = "not defined"
        elec = "not defined"
        
        if outFile is None:
            saveFile = "DMC"
            if sampleName == True:
                saveFile += f"_{samName[:6]}"
            if temperature == True:
                saveFile += "_" + str(meanTemp).replace(".","p")[:3] + "K"
            if magneticField == True:
                saveFile += "_" + mag + "T"
            if electricField == True:
                saveFile += "_" + elec + "keV"
            if fileNumber == True:
                saveFile += "_" + fileNumbers.replace(',','_')  
        else:
            saveFile = str(outFile.replace('.dat',''))

        if outFolder is None:
            outFolder = os.getcwd()

        with open(os.path.join(outFolder,saveFile)+".dat",'w') as sf:
            sf.write(fileString)

    def export_xye_format(self,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,outFolder=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

        """
        The function takes a data set and merge the files.
        Outputs a .xye file in with a comment line with info and xye data
        Saves the file with input name
        
        Kwargs:
            
            - dTheta (Float): stepsize of binning if no nins is given (default is 0.125)
            
            - twoThetaOffset (float): Linear shift of two theta, default is 0. To be used if a4 in hdf file is incorrect
            
            - Bins (list): Bins into which 2theta is to be binned (default min(2theta),max(2theta) in steps of 0.125)
            
            - outFile (str): String that will be used for outputfile. Default is automatic generated name.

            - outFolder (str): Path to folder data will be saved. Default is current working directory.
            
        - Arguments for automatic file name:
                
            - sampleName (bool): Include sample name in filename. Default is True.
        
            - temperature (bool): Include temperature in filename. Default is True.
        
            - magneticField (bool): Include magnetic field in filename. Default is False.
        
            - electricField (bool): Include electric field in filename. Default is False.
        
            - fileNumber (bool): Include sample number in filename. Default is True.
            
        Kwargs for sumDetector:

            - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
                
            - applyNormalization (bool): Use normalization files (default True)
                
            - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
            
        Returns:
            
            .xye file in with a comment line with info and xye data
        
        Note: Input is a data set.
            
        Example:
            >>> inputNumber = _tools.fileListGenerator(565,folder)
            >>> ds = DataSet.DataSet(inputNumber)
            >>> for df in ds:
            ...    if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
            ...        df.monitor = np.ones_like(df.monitor)
            >>> export_xye_format(ds)
            
        """

        twoTheta = self.twoTheta
        
        anglesMin = np.min(twoTheta)
        anglesMax = np.max(twoTheta)
        
        if bins is None:
            bins = np.arange(anglesMin-0.5*dTheta,anglesMax+0.51*dTheta,dTheta)
        
        bins,intensity,err,monitor = self.sumDetector(bins,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta)
        
        bins = bins + twoThetaOffset
        
        # find mean monitor
        meanMonitor = np.median(monitor)
        intensity[np.isnan(intensity)] = -1
        
        # rescale intensity and err
        intensity*=meanMonitor
        err*=meanMonitor
        
        step = np.mean(np.diff(bins))
        start = np.min(bins)+0.5*step
        stop = np.max(bins)-0.5*step
        
        Centres=0.5*(bins[1:]+bins[:-1])
        saveData = np.array([Centres,intensity,err])
        
        if np.all([x == self.sample[0].name for x in [s.name for s in self.sample[1:]]]):
            samName = self.sample[0].name        #.decode("utf-8")
        else:
            samName ='Unknown! Combined different sample names'

        if np.all([np.isclose(x,self.wavelength[0]) for x in self.wavelength[1:]]):
            wavelength = self.wavelength[0]
        else:
            wavelength ='Unknown! Combined different Wavelengths'
        
        temperatures = np.array([df.temperature for df in self])
        meanTemp = np.mean(temperatures)
        
        # fileNumbers = str(self.fileName) 
        # fileNumbers_short = str(int(self.fileName[0].split('n')[-1].split('.')[0]))  # 
        
        if len(self) == 1:
            year = 2022
            fileNumbers = str(int(self.fileName[0].split('n')[-1].split('.')[0]))
        else:
            year,fileNumbers = _tools.numberStringGenerator(self.fileName)
        
        titleLine1 = f"# DMC at SINQ, PSI: Sample name = {samName}, wavelength = {str(wavelength)[:5]} AA, T = {str(meanTemp)[:5]} K"
        titleLine2 = "# Filelist='dmc:{}:{}'".format(year,fileNumbers)
        titleLine3= '# '+' '.join(["{:7.3f}".format(x) for x in [start,step,stop]])+" {:7.0f}".format(meanMonitor)+'., sample="'+samName+'"'

            
        # get magnetic field
        # get electric field
        mag = "not defined"
        elec = "not defined"
        
        if outFile is None:
            saveFile = "DMC"
            if sampleName == True:
                saveFile += f"_{samName[:6]}"
            if temperature == True:
                saveFile += "_" + str(meanTemp).replace(".","p")[:3] + "K"
            if magneticField == True:
                saveFile += "_" + mag + "T"
            if electricField == True:
                saveFile += "_" + elec + "keV"
            if fileNumber == True:
                saveFile += "_" + fileNumbers.replace(',','_') 
        else:
            saveFile = str(outFile.replace('.xye',''))

        if outFolder is None:
            outFolder = os.getcwd()

        with open(os.path.join(outFolder,saveFile)+".xye",'w') as sf:
            sf.write(titleLine1+"\n")    
            sf.write(titleLine2+"\n") 
            sf.write(titleLine3+"\n") 
            np.savetxt(sf,saveData.T,delimiter='  ')
            sf.close()
        

    def updateDataFiles(self,key,value):
        if np.all([hasattr(df,key) for df in self]): # all datafiles have the key
            try:
                length = len(value)
            except TypeError:
                length = 1
            
            if length == len(self): # input has the same length as number of data files! Apply individually
                if length == 1:
                    value = [value]
                for v,df in zip(value,self):
                    setattr(df,key,v)
            elif length == 1:
                for df in self:
                    setattr(df,key,value)
            else:
                raise AttributeError('Length of DataSet is {} but received {} values for {}'.format(len(self),length,key))
                
            self._getData() # update!
            
        else:
            missing = len(self)-np.sum([hasattr(df,key) for df in self])
            if missing == 0:
                raise AttributeError('DataFiles do not contain',key)
            else:
                raise AttributeError('Not all DataFiles do not contain',key)
            
    
            
            
def add(*listinput,PSI=True,xye=True,folder=None,outFolder=None,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

    """
    
    Takes a set/series file numbers and export a added/merged file. 
    The input is read as a tuple and can be formatted as int, str, list, and several arguments separated by comma can be given. 
    If one argument is a list or str, multiple filenumbers can be given inside.
    
    Exports PSI and xye format file for all scans. 
    
    Kwargs:
        
        - listinput (tuple): The function will add/merge all elements of the tuple/list. Files can be given as int, str, list.
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - outFile (str): string for name of outfile (given without extension)

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
        
        - PSI (bool): Export PSI format. Default is True
        
        - xye (bool): Export xye format. Default is True
        
    - Arguments for automatic file name:
            
        - sampleName (bool): Include sample name in filename. Default is True.
    
        - temperature (bool): Include temperature in filename. Default is True.
    
        - magneticField (bool): Include magnetic field in filename. Default is False.
    
        - electricField (bool): Include electric field in filename. Default is False.
    
        - fileNumber (bool): Include sample number in filename. Default is True.
        
    Kwargs for sumDetector:

        - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
            
        - applyNormalization (bool): Use normalization files (default True)
            
        - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
    Example:    
        >>> add(565,566,567,(570),'571-573',[574],sampleName=False,temperature=False)
        
        output:
            DMC_565-567_570-574 as both .dat and xye files
    """    

    if folder is None:
        folder = os.getcwd()
    if outFolder is None:
        outFolder = os.getcwd()
        
    listOfDataFiles = str()
    
    if type(listinput) == tuple:
        for elemnt in listinput:
            elemnt = str(elemnt)
            elemnt = elemnt.replace('"','').replace("'","").replace('(','').replace(')','').replace('[','').replace(']','').strip(',')
            listOfDataFiles += f"{elemnt},"
        print(f"Export of added files: {listOfDataFiles[:-1]}")
        inputNumber = _tools.fileListGenerator(listOfDataFiles[:-1],folder)
        print(inputNumber)
        ds = DataSet(inputNumber)
        for df in ds:
            if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
                df.monitor = np.ones_like(df.monitor)
            print('I got ehre 2')
        if PSI == True:
            ds.export_PSI_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber)    
        if xye == True:
            ds.export_xye_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber) 
    else:
        print("Cannot export! Something wrong with input")       


# add(565,566,567,(570),'571-573',[574],sampleName=False,temperature=False)



        
def export(*listinput,PSI=True,xye=True,folder=None,outFolder=None,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

    """
    
    Takes a set file numbers and export induvidually. 
    The input is read as a tuple and can be formatted as int, str, list, and arguments separated by comma is export induvidually. 
    If one argument is a list or str, multiple filenumbers can be given inside, and they will be added/merged.
    
    Exports PSI and xye format file for all scans. 
    
    Kwargs:
        
        - listinput (tuple): the function will export all elements of the tuple/list inducidually. Files can be merged by [], '', and () notation.
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - PSI (bool): Export PSI format. Default is True
        
        - xye (bool): Export xye format. Default is True
        
        - outFile (str): string for name of outfile (given without extension)

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
        
    - Arguments for automatic file name:
            
        - sampleName (bool): Include sample name in filename. Default is True.
    
        - temperature (bool): Include temperature in filename. Default is True.
    
        - magneticField (bool): Include magnetic field in filename. Default is False.
    
        - electricField (bool): Include electric field in filename. Default is False.
    
        - fileNumber (bool): Include sample number in filename. Default is True.
        
    Kwargs for sumDetector:

        - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
            
        - applyNormalization (bool): Use normalization files (default True)
            
        - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
    Example:    
        >>> export(565,'566',[567,568,570,571],'570-573',(574,575),sampleName=None,temperature=False)  
        
        output: DMC_565, DMC_566, DMC_567_568_570_571, DMC_570-573, DMC_574-575 as both .dat and xye files
        
    """    

    if folder is None:
        folder = os.getcwd()
    if outFolder is None:
        outFolder = os.getcwd()
    
    if type(listinput) == tuple:
        for elemnt in listinput:
            elemnt = str(elemnt)
            elemnt = elemnt.replace('"','').replace("'","").replace('(','').replace(')','').replace('[','').replace(']','').strip(',')
            print(f"Export of: {elemnt}")
            inputNumber = _tools.fileListGenerator(elemnt,folder)
            try:
                ds = DataSet(inputNumber)
                for df in ds:
                    if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
                        df.monitor = np.ones_like(df.monitor)
                if PSI == True:
                    ds.export_PSI_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber)    
                if xye == True:
                    ds.export_xye_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber) 
            except:
                print(f"Cannot export! File is wrong format: {elemnt}")
    else:
        print("Cannot export! Something wrong with input")
        

# export(565,'566',[567,568,570,571],'570-573',(574,575),sampleName=False,temperature=False)  
        




def export_from(startFile,PSI=True,xye=True,folder=None,outFolder=None,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):
    
    """
    
    Takes a starting file number and export xye format file for all the following files in the folder.

    Exports PSI and xye format file for all scans. 
    
    Kwargs:
        
        - startFile (int): First file number for export
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - PSI (bool): Export PSI format. Default is True
        
        - xye (bool): Export xye format. Default is True

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
        
        - all from export_PSI_format and export_xye_format
        
    - Arguments for automatic file name:
            
        - sampleName (bool): Include sample name in filename. Default is True.
    
        - temperature (bool): Include temperature in filename. Default is True.
    
        - magneticField (bool): Include magnetic field in filename. Default is False.
    
        - electricField (bool): Include electric field in filename. Default is False.
    
        - fileNumber (bool): Include sample number in filename. Default is True.
        
    Kwargs for sumDetector:

        - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
            
        - applyNormalization (bool): Use normalization files (default True)
            
        - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
    Example:       
        >>> export_from(590,sampleName=False,temperature=False)    
        
    """
    if folder is None:
        folder = os.getcwd()
    if outFolder is None:
        outFolder = os.getcwd()

    hdf_files = [f for f in os.listdir(folder) if f.endswith('.hdf')]
    last_hdf = hdf_files[-1]

    numberOfFiles = int(last_hdf.strip('.hdf').split('n')[-1]) - int(startFile)
    
    fileList = list(range(startFile,startFile+numberOfFiles))
    
    for file in fileList:  

        file = str(file)
        file = file.replace('"','').replace("'","").replace('(','').replace(')','').replace('[','').replace(']','').replace(' ','').strip(',')
        print(f"Export of: {file}")
        try:
            inputNumber = _tools.fileListGenerator(file,folder)
            ds = DataSet(inputNumber)
            for df in ds:
                if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
                    df.monitor = np.ones_like(df.monitor)
            if PSI == True:
                ds.export_PSI_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber)    
            if xye == True:
                ds.export_xye_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber) 
        except:
            print(f"Cannot export! File is wrong format: {file}")
            
# export_from(587,sampleName=False,temperature=False)    





def export_from_to(startFile,endFile,PSI=True,xye=True,folder=None,outFolder=None,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

    """
    
    Takes a starting file number and a end file number, export for all scans between (including start and end)

    Exports PSI and xye format file for all scans. 
    
    Kwargs:
        
        - startFile (int): First file number for export
        
        - endFile (int): Final file number for export
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - PSI (bool): Export PSI format. Default is True
        
        - xye (bool): Export xye format. Default is True

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
        
        - all from export_PSI_format and export_xye_format
        
    - Arguments for automatic file name:
            
        - sampleName (bool): Include sample name in filename. Default is True.
    
        - temperature (bool): Include temperature in filename. Default is True.
    
        - magneticField (bool): Include magnetic field in filename. Default is False.
    
        - electricField (bool): Include electric field in filename. Default is False.
    
        - fileNumber (bool): Include sample number in filename. Default is True.
        
    Kwargs for sumDetector:

        - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
            
        - applyNormalization (bool): Use normalization files (default True)
            
        - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
    Example:     
        >>> export_from_to(565,570,sampleName=False,temperature=False)
        
        output: DMC_565, DMC_566, DMC_567, DMC_568, DMC_569, DMC__570 as both .xye and .dat files
        
    """
    if folder is None:
        folder = os.getcwd()
    if outFolder is None:
        outFolder = os.getcwd()

    fileList = list(range(startFile,endFile+1))
    
    for file in fileList:    
        file = str(file)
        file = file.replace('"','').replace("'","").replace('(','').replace(')','').replace('[','').replace(']','').replace(' ','').strip(',')
        print(f"Export of: {file}")
        try:
            inputNumber = _tools.fileListGenerator(file,folder)
            ds = DataSet(inputNumber)
            for df in ds:
                if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
                    df.monitor = np.ones_like(df.monitor)
            if PSI == True:
                ds.export_PSI_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber)    
            if xye == True:
                ds.export_xye_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber) 
        except:
            print(f"Cannot export! File is wrong format: {file}")


# export_from_to(565,570,sampleName=False,temperature=False)







def export_list(listinput,PSI=True,xye=True,folder=None,outFolder=None,dTheta=0.125,twoThetaOffset=0,bins=None,outFile=None,applyNormalization=True,correctedTwoTheta=True,sampleName=True,temperature=False,magneticField=False,electricField=False,fileNumber=False):

    """
    
    Takes a list and export all elements induvidually. If a list is given inside the list, these files will be added/merged.

    Exports PSI and xye format file for all scans. 
    
    Kwargs:
        
        - list input (list: List of files that will be exported.
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - PSI (bool): Export PSI format. Default is True
        
        - xye (bool): Export xye format. Default is True

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
        
        - all from export_PSI_format and export_xye_format
        
    - Arguments for automatic file name:
            
        - sampleName (bool): Include sample name in filename. Default is True.
    
        - temperature (bool): Include temperature in filename. Default is True.
    
        - magneticField (bool): Include magnetic field in filename. Default is False.
    
        - electricField (bool): Include electric field in filename. Default is False.
    
        - fileNumber (bool): Include sample number in filename. Default is True.
        
    Kwargs for sumDetector:

        - twoThetaBins (array): Actual bins used for binning (default [min(twoTheta)-dTheta/2,max(twoTheta)+dTheta/2] in steps of dTheta=0.125 Deg)
            
        - applyNormalization (bool): Use normalization files (default True)
            
        - correctedTwoTheta (bool): Use corrected two theta for 2D data (default true)
        
    Example:    
        >>> export_list([565,566,567,[569,570]],sampleName=False,temperature=False) 
        
        output: DMC_565, DMC_566, DMC_567, DMC_569_570 as both .xye and .dat files
        
    """    

    if folder is None:
        folder = os.getcwd()
    if outFolder is None:
        outFolder = os.getcwd()
        
    for file in listinput:   
        file = str(file)
        file = file.replace('"','').replace("'","").replace('(','').replace(')','').replace('[','').replace(']','').replace(' ','').strip(',')
        print(f"Export of: {file}")           
        try:
            inputNumber = _tools.fileListGenerator(file,folder)
            ds = DataSet(inputNumber)
            for df in ds:
                if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
                    df.monitor = np.ones_like(df.monitor)
            if PSI == True:
                ds.export_PSI_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber)    
            if xye == True:
                ds.export_xye_format(dTheta=dTheta,twoThetaOffset=twoThetaOffset,bins=bins,outFile=outFile,outFolder=outFolder,applyNormalization=applyNormalization,correctedTwoTheta=correctedTwoTheta,sampleName=sampleName,temperature=temperature,magneticField=magneticField,electricField=electricField,fileNumber=fileNumber) 
        except:
            print(f"Cannot export! File is wrong format: {file}")
            

# export_list([565,566,567,[569,570]],sampleName=False,temperature=False)            
                

                
def subtract_PSI(file1,file2,outFile=None,folder=None,outFolder=None):

    """
    
    This function takes two .dat files in PSI format and export a differnce curve with correct uncertainties. 
    
    The second file is scaled after the monitor of the first file.
    
    Kwargs:
        
        - PSI (bool): Subtract PSI format. Default is True
        
        - xye (bool): Subtract xye format. Default is True

        - folder (str): Path to directory for data files, default is current working directory
        
        - outFile (str): string for name of outfile (given without extension)

        - outFolder (str): Path to folder data will be saved. Default is current working directory.
                
    Example:
        >>> subtract('DMC_565.dat','DMC_573')
    
    """

    if folder is None:
        folder = os.getcwd()
        
    with open(os.path.join(folder,file1.replace('.dat','')+'.dat'),'r') as rf:
        allinfo1 = rf.readlines()
        rf.close()

    with open(os.path.join(folder,file2.replace('.dat','')+'.dat'),'r') as rf:
        allinfo2 = rf.readlines()
        rf.close()    

    info1 = allinfo1[:3] 
    info2 = allinfo2[:3] 

    if info1[2].split(',')[0].split(',')[0] != info2[2].split(',')[0].split(',')[0]:
        return print('Not same range of files! Cannot subtract.')          
        
    infoStr1 = (info1[2].split(',')[0].strip('#').replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('  ',' '))
    infoArr1 = [float(x) for x in infoStr1[1:].split(' ')]

    infoStr2 = (info2[2].split(',')[0].strip('#').replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('  ',' '))
    infoArr2 = [float(x) for x in infoStr2[1:].split(' ')]
        
    monitor1 = infoArr1[3]
    monitor2 = infoArr2[3]
    monitorRatio = monitor1/monitor2    
    dataPoints = int((infoArr1[2]-infoArr1[0]) / infoArr1[1]) + 1
    subInt = []
    subErr = []
    
    dataLines = int(np.ceil(dataPoints/10)) 
    commentlines = 3
    
    for intLines in range(dataLines): 
        subIntList = []
        subErrList = []
        intline1= allinfo1[intLines+commentlines]
        intline2= allinfo2[intLines+commentlines]
        errline1= allinfo1[intLines+dataLines+commentlines]
        errline2= allinfo2[intLines+dataLines+commentlines]
        intensity1 = [float(x) for x in intline1[:-2].replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('nan','').replace('na','').split(' ') if x != '' ]  
        intensity2 = [float(x) for x in intline2[:-2].replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('nan','').replace('na','').split(' ') if x != '' ] 
        err1 = [float(x) for x in errline1[:-2].replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('nan','').replace('na','').split(' ') if x != '' ] 
        err2 = [float(x) for x in errline2[:-2].replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('nan','').replace('na','').split(' ') if x != '' ] 
        for i, j in zip(intensity1,intensity2):
            subIntList.append(i-j*monitorRatio)
        for h, k in zip(err1,err2):
            subErrList.append(np.sqrt(h**2 + monitorRatio**2 * k**2))
        subInt.append(subIntList)
        subErr.append(subErrList)
    
    titleLine = str(info1[0]).strip('\n') + ', subtracted: ' + str(info2[0])  + str(info1[1]) + str(info1[2]).strip('\n')
    dataLinesInt = '\n'.join([' '+' '.join(["{:6.0f}.".format(x) for x in line]) for line in subInt])
    dataLinesErr = '\n'.join([' '+' '.join(["{:7.1f}".format(x) for x in line]) for line in subErr])
    indexParamLines = dataLines*2 + commentlines
    paramLine1 = '\n'.join([str(line).strip('\n') for line in allinfo1[indexParamLines:]])
    paramLine2 = ' subtracted:'
    paramLine3 = str(info2[0])  + str(info1[1]) + str(info1[2]).strip('\n')
    paramLine4 = ''.join([str(line) for line in allinfo2[indexParamLines:]])
    fileString = '\n'.join([titleLine,dataLinesInt,dataLinesErr,paramLine1,paramLine2,paramLine3,paramLine4])
    
    if outFile is None:
        saveFile = file1.replace('.dat','') + '_sub_' + file2.replace('.dat','')
    else:
        saveFile = str(outFile.replace('.dat',''))

    print(f'Subtracting PSI: {file1}.dat minus {file2}.dat') 

    if outFolder is None:
        outFolder = os.getcwd()

    with open(os.path.join(outFolder,saveFile)+".dat",'w') as sf:
        sf.write(fileString)
    
    
    
# subtract_PSI('DMC_565','DMC_573')


def subtract_xye(file1,file2,outFile=None,folder=None,outFolder=None):

    """
    
    This function takes two .xye files and export a differnce curve with correct uncertainties. 
    
    The second file is scaled after the monitor of the first file.
    
    Kwargs:
        
        - PSI (bool): Subtract PSI format. Default is True
        
        - xye (bool): Subtract xye format. Default is True

        - folder (str): Path to directory for data files, default is current working directory
        
        - outFile (str): string for name of outfile (given without extension)

        - outFolder (str): Path to folder data will be saved. Default is current working directory. 
        
    Example:
        >>> subtract('DMC_565.xye','DMC_573')
        
    """
    
    if folder is None:
        folder = os.getcwd()
    
    data1 = np.genfromtxt(os.path.join(folder,file1.replace('.xye','')+'.xye'), delimiter='  ')
    data2 = np.genfromtxt(os.path.join(folder,file2.replace('.xye','')+'.xye'), delimiter='  ')  
    
    with open(os.path.join(folder,file1.replace('.xye','')+'.xye'),'r') as rf:
        info1 = rf.readlines()[:3]
        rf.close()

    with open(os.path.join(folder,file2.replace('.xye','')+'.xye'),'r') as rf:
        info2 = rf.readlines()[:3]
        rf.close()

    if info1[2].split(',')[0].split(',')[0] != info2[2].split(',')[0].split(',')[0]:
        return print('Not same range of files! Cannot subtract.')          
        
    infoStr1 = (info1[2].split(',')[0].strip('#').replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('  ',' '))
    infoArr1 = [float(x) for x in infoStr1[1:].split(' ')]

    infoStr2 = (info2[2].split(',')[0].strip('#').replace('  ',' ').replace('  ',' ').replace('  ',' ').replace('  ',' '))
    infoArr2 = [float(x) for x in infoStr2[1:].split(' ')]

    monitorRatio = infoArr1[3]/infoArr2[3]

    subInt = np.subtract(data1[:,1], np.multiply(monitorRatio,(data2[:,1])))

    intErr2 = monitorRatio * data2[:,2]
    
    subErr = 0 * data2[:,1]
    
    for i in range(len(data1[:,2])):
        subErr[i] = np.sqrt( (data1[i,2])**2 + (intErr2[i])**2 ) 
    
    saveData = np.array([data1[:,0],subInt,subErr])

    if outFile is None:
        saveFile = file1.replace('.xye','') + '_sub_' + file2.replace('.xye','')
    else:
        saveFile = str(outFile.replace('.xye',''))

    print(f'Subtracting xye: {file1}.xye minus {file2}.xye')    

    if outFolder is None:
        outFolder = os.getcwd()

    with open(os.path.join(outFolder,saveFile)+".xye",'w') as sf:
        sf.write('# ' + str(info1) + "\n")   
        sf.write("# subtracted file: \n") 
        sf.write('# ' + str(info2) + "\n") 
        np.savetxt(sf,saveData.T,delimiter='  ')
        sf.close()

        
    # subtract_xye('DMC_565','DMC_573')

def subtract(file1,file2,PSI=True,xye=True,outFile=None,folder=None,outFolder=None):
    """

    This function takes two files and export a differnce curve with correct uncertainties. 

    The second file is scaled after the monitor of the first file.

    Kwargs:
        
        - PSI (bool): Subtract PSI format. Default is True
        
        - xye (bool): Subtract xye format. Default is True
        
        - folder (str): Path to directory for data files, default is current working directory
        
        - outFile (str): string for name of outfile (given without extension)

        - outFolder (str): Path to folder data will be saved. Default is current working directory. 
        
    Example:
        >>> subtract('DMC_565.xye','DMC_573')

    """


    if folder is None:
        folder = os.getcwd()

    file1 = file1.replace('.xye','').replace('.dat','')
    file2 = file2.replace('.xye','').replace('.dat','')

    if PSI == True:
        try:
            subtract_PSI(file1,file2,outFile,folder=folder,outFolder=outFolder)
        except:
            print('Cannot subtract PSI format files')
    if xye == True:
        try:
            subtract_xye(file1,file2,outFile,folder=folder,outFolder=outFolder)
        except:
            print('Cannot subtract xye format files')
        
        
#subtract('DMC_565.xye','DMC_573')

def sort_export():
    
    pass

def export_help(): 
    print(" ")
    print(" The following commands are avaliable for export of powder data in DMCpy:")
    print(" ")
    print("     export, add, export_from, export_from_to, export_list")
    print(" ")
    print(" They export both PSI and xye format by default. Can be deactivated by the arguments PSI=False and xye=False")
    print(" ")
    print('      - export: For export of induvidual sets of scan files. Files can be merged by [] or "" notation, i.e. list or strings.')
    print('      - add: TThe function adds/merge all the files given. ')
    print("      - export_from: For export of all data files in a folder after a startfile")
    print("      - export_from_to: It exports all files between and including two given files")
    print("      - export_list: Takes a list and export all the files separatly. If a list is given inside the list, the files will be merged ")
    print(" ")
    print(" Examples:  ")
    print('      >>> export(565,"566",[567,568,570,571],"570-573",(574,575),sampleName=False,temperature=False)  ')
    print("      >>> add(565,566,567,(570),'571-573',[574],sampleName=False,temperature=False) ")
    print("      >>> export_from(590,fileNumber=True)  ")
    print("      >>> export_from_to(565,570,dTheta=0.25,twoThetaOffset=2.0)")
    print("      >>> export_list([565,566,567,570],temperature=False,xye=False)")
    print("      >>> export_list([565,566,567,[568,569,570]]) # This is an example of list inside a list. 568,569,570 will be merged in this case. ")
    print('      >>> add("565,567,570-573",outFile="mergefilename")')
    print('      >>> export(565,folder=r"Path\To\Data\Folder")   #Note r"..." notation')      
    print(" ")
    print(" ")
    print(" Most important kewords and aguments:")
    print(" ")
    print("     - dTheta (float): stepsize of binning if no bins is given (default is 0.125)")
    print("     - outFile (str): String that will be used for outputfile. Default is automatic generated name.")
    print("     - outFolder (str): Path to folder data will be saved. Default is current working directory.")
    print("     - twoThetaOffset (float): Linear shift of two theta, default is 0. To be used if a4 in hdf file is incorrect")
    print(" ")
    print(" Arguments for automatic file name:")
    print(" ")
    print("     - sampleName (bool): Include sample name in filename. Default is True.")
    print("     - temperature (bool): Include temperature in filename. Default is True.")
    print("     - fileNumber (bool): Include sample number in filename. Default is True.")
    print("     - magneticField (bool): Include magnetic field in filename. Default is False.")
    print("     - electricField (bool): Include electric field in filename. Default is False.")
    print(" ")
    print(" ")
    print(" There is also a subtract function for subtracting PSI format files and xye format files. ")    
    print(" The files are normalized to the onitor of the first dataset.")
    print(" Input is two existing filenames with or without extenstion. ")
    print(" PSI and xye format can be deactivated by PSI = False and xye = False")    
    print(" Alternatively can subtract_PSI or subtract_xye be used")
    print(" ")
    print("      >>> subtract('DMC_565.xye','DMC_573')")    
    print(" ")
    

from matplotlib.ticker import FuncFormatter

def generate1DAxis(q1,q2,rlu=True,outputFunction=print):
    fig,ax = plt.subplots()
    ax = plt.gca()
    q1 = np.asarray(q1,dtype=float)
    q2 = np.asarray(q2,dtype=float)
    
    if rlu:
        variables = ['H','K','L']
    else:
        variables = ['Qx','Qy','Qz']
    
        # Start points defined form cut
    ax.startPoint = q1
    ax.endPoint = q2
    
    
    # plot direction is q2-q1, without normalization making all x points between 0 and 1
    
    ax.plotDirection = _tools.LengthOrder(np.array(q2-q1)).reshape(-1,1)
    
    # Calculate the needed precision for x-axis plot
    def calculateXPrecision(ax):
        # Find diff for current view
        diffPlotPosition = np.diff(ax.get_xlim())[0]
        diffAlongPlot = ax.plotDirection*diffPlotPosition
        
        numTicks = len(ax.xaxis.get_ticklocs())
        
        # take the smallest value which is chaning (i.e. is along the plot direction)
        minChange = np.min(np.abs(diffAlongPlot[ax.plotDirection.T.flatten()!=0])) /numTicks 
        
        
        # find the largest integer closest to the wanted precision
        ax.set_precision(int(-np.floor(np.log10(minChange)))+1)
    

    def calculateIndex(binDistance,x):
        idx = np.argmin(np.abs(binDistance-x))
        return idx
    
    def calculatePosition(ax,x):
        if isinstance(x,(np.ndarray)):
            return (x*ax.plotDirection.T+ax.startPoint)
        else:
            return (x*ax.plotDirection.T+ax.startPoint).flatten()
    
    def calculatePositionInv(ax,h,k,l):
        HKL = np.asarray([h,k,l])
        return np.dot((HKL-ax.startPoint.reshape(3,1)).T,ax.plotDirection)/(np.dot(ax.plotDirection.T,ax.plotDirection)).T
        # return np.dot((HKL-ax.startPoint.reshape(3,1)).T,ax.plotDirection.reshape(3,1))/(np.dot(ax.plotDirection.T,ax.plotDirection))
    
    # Add methods to the axis
    
    

    ax._x_precision = 2
    ax.fmtPrecisionString = '{:.'+str(2)+'f}'
    # Dynamic add setter and getter to ax.precision
    
    def set_precision(ax,value):
        ax._x_precision = value
        ax.fmtPrecisionString = '{:.'+str(ax._x_precision)+'f}'
        ax.get_figure().tight_layout()
        
    
    
    ax.calculatePosition = lambda x: calculatePosition(ax,x)
    ax.calculatePositionInv = lambda h,k,l: calculatePositionInv(ax,h,k,l)
    
    ax.calculateIndex = lambda x: calculateIndex(ax.Data['binDistance'],x)
    ax.calculateXPrecision = calculateXPrecision
    ax.set_precision = lambda value: set_precision(ax,value)
    ax.calculateXPrecision(ax)
    
    # Format the x label as well as the format_coord
    if rlu==False:
        xlabel = r'[$Q_x [\AA^{-1}]$, $Q_y [\AA^{-1}]$, $Q_z [\AA^{-1}]$]'
        
        ax.set_xlabel(xlabel)
        def format_coord(x,y,ax):# pragma: no cover
            qx,qy,qz = ax.calculatePosition(x)
            return  "qx = {0:.3e}, qy = {1:.3e}, qz = {2:.3e}, I = {3:0.4e}".format(qx,qy,qz,y)
    else:
        xlabel = '[$Q_h$ [RLU], $Q_k$ [RLU], $Q_l$ [RLU]]'
        ax.set_xlabel(xlabel)
        
        def format_coord(x,y,ax):# pragma: no cover
            h,k,l = ax.calculatePosition(x)
            return  "H = {0:.3e}, K = {1:.3e}, L = {2:.3e}, I = {3:0.4e}".format(h,k,l,y)
        
    
    # Create a custom major formatter to show the multi-D position on the x-axis
    def major_formatter(ax,tickPosition,tickNumber):
        positions = list(ax.calculatePosition(tickPosition))
        return '\n'.join([ax.fmtPrecisionString.format(pos) for pos in positions])
    
    ax.xaxis.set_major_formatter(FuncFormatter(lambda x,i: major_formatter(ax,x,i)))
    
    # Create the onclick behaviour
    def onclick(event,ax,outputFunction):# pragma: no cover
        if ax.in_axes(event):
            try:
                C = ax.get_figure().canvas.cursor().shape() # Only works for pyQt5 backend
            except:
                pass
            else:
                if C != 0: # Cursor corresponds to arrow
                    return
    
            x = event.xdata
            y = event.ydata
            printString = ax.format_coord(x,y)
            
            outputFunction(printString)
    
    
    # connect methods
    ax.format_coord = lambda x,y: format_coord(x,y,ax)
    ax._button_press_event = ax.figure.canvas.mpl_connect('button_press_event',lambda event:onclick(event,ax,outputFunction=outputFunction))
    
    ax.callbacks.connect('xlim_changed',ax.calculateXPrecision)
    # Make the layouyt fit
    ax.get_figure().tight_layout()

    return ax


import sys 
import os
# sys.path.append('/home/lass/Dropbox/PhD/Software/DMCpy/')

from Tutorial_Class import Tutorial


def Tester():
    import numpy as np
    import matplotlib.pyplot as plt

    fig,ax = plt.subplots()
    fig.savefig(os.path.join(os.getcwd(),r'docs/Tutorials/TEST.png'),format='png',dpi=300)
    
title = 'Test of Tutorials!'

introText = 'This is a simple test script to check that all is working as expected. Below should be an image of an empty axis...'\
+' if not, then Houston we have a problem '

outroText = 'Showcasing functionality\n'+len('Showcasing functionality')*'-'+'\nThis should be the showcasing...'\
+'An example is shown in the above code generating the figure below:'\
+'\n .. figure:: TEST.png\n  :width: 50%\n  :align: center\n\n'

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('Testing',introText,outroText,Tester,fileLocation = os.path.join(os.getcwd(),r'docs/Tutorials'))

def test_Testing():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()